# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

import pytest

from capellambse.filehandler import memory


def test_MemoryFileHandler_raises_ValueError_for_invalid_path():
    with pytest.raises(ValueError):
        memory.MemoryFileHandler(path="memory://invalid")


def test_MemoryFileHandler_raises_FileNotFoundError_for_nonexistent_file():
    fh = memory.MemoryFileHandler()
    with pytest.raises(FileNotFoundError):
        fh.open("test.txt")


def test_MemoryFileHandler_preserves_written_data():
    fh = memory.MemoryFileHandler()
    with fh.open("test.txt", "w") as f:
        f.write(b"Hello, World!")

    with fh.open("test.txt", "r") as f:
        assert f.read() == b"Hello, World!"


def test_MemoryFiles_return_bytes_objects_from_read():
    fh = memory.MemoryFileHandler()
    with fh.open("test.txt", "w") as f:
        f.write(b"Hello, World!")

    with fh.open("test.txt", "r") as f:
        assert isinstance(f.read(), bytes)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Helpers for working with models in CLI scripts."""
from __future__ import annotations

__all__ = ["ModelCLI", "enumerate_known_models", "loadcli"]

import collections.abc as cabc
import importlib.abc
import importlib.resources as imr
import json
import os
import typing as t

import capellambse

try:
    import click

    class ModelCLI(click.ParamType):
        """Declare an option that loads a model.

        Use instances of this class for the ``type=`` argument to
        ``@click.option`` etc.

        Examples
        --------
        .. code-block:: python

           @click.command()
           @click.option("-m", "--model", type=capellambse.ModelCLI())
           def main(model: capellambse.MelodyModel) -> None:
               ...
        """

        name = "CAPELLA_MODEL"

        def convert(self, value: t.Any, param, ctx) -> capellambse.MelodyModel:
            if isinstance(value, capellambse.MelodyModel):
                return value

            try:
                return loadcli(value)
            except ValueError as err:
                self.fail(err.args[0], param, ctx)
                assert False

except ImportError:

    def ModelCLI(*__, **_):  # type: ignore[no-redef]
        """Raise a dependency error."""
        raise RuntimeError("click is not installed")


def enumerate_known_models() -> cabc.Iterator[importlib.abc.Traversable]:
    """Enumerate the models that are found in the ``known_models`` folders.

    Two places are searched for models: The *known_models* folder in the
    user's configuration directory, and the *known_models* folder in the
    installed ``capellambse`` package.

    In order to make a custom model known, place a JSON file in one of
    these *known_models* folders. It should contain a dictionary with
    the keyword arguments to :class:`capellambse.MelodyModel` -
    specifically it needs a ``path``, an ``entrypoint``, and any
    additional arguments that the underlying
    :class:`capellambse.filehandler.FileHandler` might need to gain
    access to the model.

    Files in the user's configuration directory take precedence over
    files in the package directory. If a file with the same name exists
    in both places, the one in the user's configuration directory will
    be used.

    Be aware that relative paths in the JSON will be interpreted
    relative to the current working directory.
    """
    names = set[str]()
    userdir = capellambse.dirs.user_config_path.joinpath("known_models")
    configdir = imr.files(capellambse).joinpath("known_models")
    i: importlib.abc.Traversable
    try:
        for i in userdir.iterdir():
            if i.name.endswith(".json") and i.is_file():
                names.add(i.name)
                yield i
    except FileNotFoundError:
        pass
    for i in configdir.iterdir():
        if i.name.endswith(".json") and i.is_file() and i.name not in names:
            yield i


def loadcli(value: str | os.PathLike[str]) -> capellambse.MelodyModel:
    """Load a model from a file or JSON string.

    This function tries to load a model from the following inputs:

    - A str or PathLike pointing to an ``.aird`` file
    - A str or PathLike pointing to a ``.json`` file, which contains the
      arguments to instantiate a :class:`capellambse.MelodyModel`
    - The contents of such a JSON file (as string)

    See Also
    --------
    capellambse.cli_helpers.CLIModel :
        This function wrapped as a ``click.ParamType``
    """
    if isinstance(value, str):
        if value.endswith((".aird", ".json")):
            return _load_from_file(value)
        try:
            data = json.loads(value)
        except json.JSONDecodeError:
            return _load_from_file(value)
        return capellambse.MelodyModel(**data)

    if isinstance(value, os.PathLike):
        value = os.fspath(value)
        if isinstance(value, str):
            return _load_from_file(value)

    raise TypeError("value must be a str or PathLike returning str")


def _load_from_file(value: str) -> capellambse.MelodyModel:
    if value.endswith(".aird"):
        return capellambse.MelodyModel(value)

    if value.endswith(".json"):
        with open(value, encoding="utf-8") as file:
            data = json.load(file)
    else:
        value = value + ".json"
        for known in enumerate_known_models():
            if value == known.name:
                data = json.loads(known.read_text(encoding="utf-8"))
                break
        else:
            try:
                data = json.loads(value)
            except json.JSONDecodeError:
                raise ValueError(
                    "value is not a known model nor contains valid JSON"
                ) from None
    return capellambse.MelodyModel(**data)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Module to populate the diagram cache with native Capella diagrams."""

from __future__ import annotations

import collections
import datetime
import itertools
import json
import logging
import pathlib
import re
import shutil
import typing as t

import lxml
from lxml import etree

import capellambse.model.diagram
from capellambse import _native
from capellambse.filehandler import local
from capellambse.model import modeltypes

E = lxml.builder.ElementMaker()
BAD_FILENAMES = frozenset(
    {"AUX", "CON", "NUL", "PRN"}
    | {f"COM{i}" for i in range(1, 10)}
    | {f"LPT{i}" for i in range(1, 10)}
)
LOGGER = logging.getLogger(__name__)
VALID_FORMATS = frozenset(
    {
        "bmp",
        "gif",
        "jpg",
        "png",
        "svg",
    }
)
VIEWPOINT_ORDER = (
    "Common",
    "Operational Analysis",
    "System Analysis",
    "Logical Architecture",
    "Physical Architecture",
)


class _IndexEntry(t.TypedDict):
    uuid: str
    name: str
    type: modeltypes.DiagramType
    viewpoint: str
    success: bool


def export(
    capella: str,
    model: capellambse.MelodyModel,
    *,
    format: str,
    index: bool,
    force: t.Literal["exe", "docker", None],
) -> None:
    if not isinstance(
        getattr(model, "_diagram_cache", None), local.LocalFileHandler
    ):
        raise TypeError(
            "Diagram cache updates are only supported for local paths"
        )

    format = format.lower()
    if format not in VALID_FORMATS:
        supported = ", ".join(sorted(VALID_FORMATS))
        raise ValueError(
            f"Invalid image format {format!r}, supported are {supported}"
        )

    diag_cache_dir = pathlib.Path(model._diagram_cache.path)
    try:
        shutil.rmtree(diag_cache_dir)
    except FileNotFoundError:
        pass
    diag_cache_dir.mkdir(parents=True)

    native_args = _find_executor(model, capella, force)
    with _native.native_capella(model, **native_args) as cli:
        assert cli.project
        cli(
            *_native.ARGS_CMDLINE,
            "-appid",
            "org.polarsys.capella.exportRepresentations",
            "-imageFormat",
            format.upper(),
            "-outputfolder",
            "/main_model/output",
            "-forceoutputfoldercreation",
        )

        diagrams = _copy_images(
            model,
            cli.project / "main_model" / "output",
            diag_cache_dir,
            format,
        )
        if index:
            _write_index(model, format, diag_cache_dir, diagrams)


def _find_executor(
    model: capellambse.MelodyModel,
    capella: str,
    force: t.Literal["exe", "docker", None],
) -> dict[str, str]:
    assert model.info.capella_version
    capella = capella.replace("{VERSION}", model.info.capella_version)
    native_args: dict[str, str] = {}
    if force == "docker":
        native_args["docker"] = capella
    elif force == "exe":
        native_args["exe"] = capella
    else:
        if pathlib.Path(capella).is_absolute():
            native_args["exe"] = capella
        elif pathlib.Path(capella).parent == pathlib.Path("."):
            exe = shutil.which(capella)
            if exe:
                native_args["exe"] = exe
            else:
                raise ValueError(f"Not found in PATH: {capella}")
        else:
            native_args["docker"] = capella
    assert "exe" in native_args or "docker" in native_args
    return native_args


def _copy_images(
    model: capellambse.MelodyModel,
    srcdir: pathlib.Path,
    destdir: pathlib.Path,
    extension: str,
) -> list[_IndexEntry]:
    name_counts = collections.defaultdict[str, int](lambda: -1)
    index: list[_IndexEntry] = []
    files = {i.name: i for i in srcdir.glob("**/*") if i.is_file()}
    copy = (shutil.copyfile, _copy_and_sanitize_svg)[extension == "svg"]

    for i in model.diagrams:
        entry: _IndexEntry = {
            "uuid": i.uuid,
            "name": i.name,
            "type": i.type,
            "viewpoint": i.viewpoint,
            "success": False,
        }
        index.append(entry)

        name_counts[i.name] = c = name_counts[i.name] + 1
        if c == 0:
            name = _sanitize_filename(i.name + f".{extension}")
        else:
            name = _sanitize_filename(i.name + f"_{c}.{extension}")

        if name not in files:
            continue

        try:
            copy(srcdir / files[name], destdir / f"{i.uuid}.{extension}")
        except Exception:
            LOGGER.exception("Cannot copy diagram %s (%s)", i.name, i.uuid)
        else:
            entry["success"] = True

    return index


def _sanitize_filename(fname: str) -> str:
    r"""Sanitize the filename.

    This function removes all characters that are illegal in file
    names on Windows operating systems, and prefixes reserved names
    with an underscore.

    Note that this also includes the two common directory separators
    (``/`` and ``\``).

    Notes
    -----
    Refer to the `MSDN article about file names`__ for more details.

    __ https://docs.microsoft.com/en-us/windows/win32/fileio/naming-a-file?redirectedfrom=MSDN#naming-conventions
    """
    fname = fname.rstrip(" .")
    fname = re.sub(
        '[\x00-\x1f<>:"/\\\\|?*]',
        lambda m: "-"[ord(m.group(0)) < 32 :],
        fname,
    )
    if fname.split(".")[0].upper() in BAD_FILENAMES:
        fname = f"_{fname}"
    return fname


def _copy_and_sanitize_svg(src: pathlib.Path, dest: pathlib.Path) -> None:
    """Copy ``src`` to ``dest`` and post process SVG diagram.

    Post-processing stops propagation of default ``fill`` and ``stroke``
    styling into elements that don't have these stylings. Fixates
    ``font-family`` to ``'Open Sans','Segoe UI',Arial,sans-serif`` and
    deletes ``stroke- miterlimit``.
    """
    tree = etree.parse(src)
    for elm in tree.iter():
        attrib = elm.attrib
        if "font-family" in attrib:
            attrib["font-family"] = "'Open Sans','Segoe UI',Arial,sans-serif"
        if "stroke-miterlimit" in attrib:
            del attrib["stroke-miterlimit"]
        if elm.tag == "{http://www.w3.org/2000/svg}clipPath":
            attrib.update({"fill": "transparent", "stroke": "transparent"})

    svg = etree.tostring(tree, pretty_print=True)
    dest.write_bytes(svg)


def _write_index(
    model: capellambse.MelodyModel,
    extension: str,
    dest: pathlib.Path,
    index: list[_IndexEntry],
) -> None:
    now = datetime.datetime.now().strftime("%A, %Y-%m-%d %H:%M:%S")
    title = f"Capella diagram cache for {model.name!r}"
    html = E.html(
        E.head(
            E.style(
                "a:active, a:hover, a:link, a:visited {text-decoration: none}"
                " .missing {color: red}"
                " .helptext {text-decoration: dashed underline; cursor: help}"
                " .small {font-size: 60%}"
                " .uuid {color: #AAA; font-size: 60%}"
            )
        ),
        body := E.body(E.h1(title), E.p({"class": "small"}, "Created: ", now)),
    )

    def sortkey(entry: _IndexEntry):
        try:
            vp_index = VIEWPOINT_ORDER.index(entry["viewpoint"])
        except ValueError:
            vp_index = len(VIEWPOINT_ORDER)
        return (vp_index, entry["viewpoint"], entry["name"])

    diagrams = sorted(index, key=sortkey)
    for vp, diags in itertools.groupby(diagrams, key=lambda i: i["viewpoint"]):
        body.append(E.h2(vp))
        body.append(ol := E.ol())
        for diagram in diags:
            uuid = E.span(diagram["uuid"], {"class": "uuid"})
            tlabel = E.span(
                {"title": diagram["type"].value, "class": "helptext"},
                f"[{diagram['type'].name}]",
            )
            if diagram["success"]:
                href = f"{diagram['uuid']}.{extension}"
                label = E.a({"href": href}, diagram["name"])
            else:
                label = E.span({"class": "missing"}, diagram["name"])
            ol.append(E.li(tlabel, " ", label, " ", uuid))

    dest.joinpath("index.json").write_text(json.dumps(index, cls=IndexEncoder))
    dest.joinpath("index.html").write_bytes(etree.tostring(html))


class IndexEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, modeltypes.DiagramType):
            return o.name
        return super().default(o)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Implements a high-level interface to Capella projects."""
from __future__ import annotations

__all__ = [
    "ElementList",
    "GenericElement",
    "NonUniqueMemberError",
    "MelodyModel",
]

import collections.abc as cabc
import logging
import os
import pathlib
import re
import shutil
import typing as t

from lxml import etree

import capellambse
import capellambse.helpers
import capellambse.pvmt
from capellambse import _diagram_cache, filehandler, loader

from . import common, diagram  # isort:skip

# Architectural Layers
from .crosslayer import (  # isort:skip
    capellacommon,
    capellacore,
    cs,
    fa,
    information,
    interaction,
    modellingcore,
)
from .layers import oa, ctx, la, pa  # isort:skip

# Exports
from .common import (  # isort:skip
    ElementList,
    GenericElement,
    ModelObject,
    NonUniqueMemberError,
)

LOGGER = logging.getLogger(__name__)
XT_PROJECT = "org.polarsys.capella.core.data.capellamodeller:Project"
XT_LIBRARY = "org.polarsys.capella.core.data.capellamodeller:Library"
XT_SYSENG = "org.polarsys.capella.core.data.capellamodeller:SystemEngineering"


@common.xtype_handler(None, XT_PROJECT, XT_LIBRARY)
class MelodyModel:
    """Provides high-level access to a model.

    This class builds upon the lower-level
    :class:`~capellambse.loader.core.MelodyLoader` to provide an
    abstract, high-level interface for easy access to various model
    aspects.
    """

    oa = common.DirectProxyAccessor(oa.OperationalAnalysis, rootelem=XT_SYSENG)
    sa = common.DirectProxyAccessor(ctx.SystemAnalysis, rootelem=XT_SYSENG)
    la = common.DirectProxyAccessor(la.LogicalArchitecture, rootelem=XT_SYSENG)
    pa = common.DirectProxyAccessor(
        pa.PhysicalArchitecture, rootelem=XT_SYSENG
    )
    enumeration_property_types = common.DirectProxyAccessor(
        capellacore.EnumerationPropertyType, aslist=common.ElementList
    )
    property_value_packages = common.DirectProxyAccessor(
        capellacore.PropertyValuePkg, aslist=common.ElementList
    )
    diagrams = diagram.DiagramAccessor(
        None, cacheattr="_MelodyModel__diagram_cache"
    )

    uuid = common.AttributeProperty(
        "id",
        writable=False,
        __doc__="The unique ID of the model's root element.",
    )
    name = common.AttributeProperty("name", __doc__="The name of this model.")

    _diagram_cache: filehandler.FileHandler
    _diagram_cache_subdir: pathlib.PurePosixPath
    _constructed: bool

    def __init__(
        self,
        path: str | os.PathLike,
        *,
        diagram_cache: (
            str
            | os.PathLike
            | filehandler.FileHandler
            | dict[str, t.Any]
            | None
        ) = None,
        diagram_cache_subdir: str | pathlib.PurePosixPath | None = None,
        jupyter_untrusted: bool = False,
        fallback_render_aird: bool = False,
        **kwargs: t.Any,
    ) -> None:
        """Load a project.

        For complete information on which exact ``kwargs`` are
        supported, consult the documentation of the used file handler.
        Refer to the "See Also" section for a collection of links.

        Below are some common parameter names and their usual meanings.
        Not all file handlers support all parameters.

        .. note::
            Passing in arguments that are not accepted by the selected
            file handler will result in an exception being raised.
            Similarly, leaving out arguments that are required by the
            file handler will also result in an exception.

        Parameters
        ----------
        path
            Path or URL to the project. The following formats are
            accepted:

            * A path to a local ``.aird`` file.
            * A path to a local directory (requires ``entrypoint``).
            * An SCP-style short URL, which will be treated as referring
              to a Git repository.

              Example: ``git@github.com:DSD-DBS/py-capellambse.git``
            * A remote URL, with a protocol or prefix that indicates
              which file handler to invoke (requires ``entrypoint``).

              Some examples:

              * ``git://git.example.com/model/coffeemaker.git``
              * ``git+https://git.example.com/model/coffeemaker.git``
              * ``git+ssh://git@git.example.com/model/coffeemaker.git``

              .. note:: Depending on the exact file handler, saving back
                 to a remote location might fail with ``update_cache``
                 set to ``False``. See :meth:`save` for more details.
        entrypoint: str
            Entrypoint from path to the main ``.aird`` file.
        revision: str
            The revision to use, if loading a model from a version
            control system like git. Defaults to the current HEAD. If
            the used VCS does not have a notion of "current HEAD", this
            argument is mandatory.
        disable_cache: bool
            Disable local caching of remote content.
        update_cache: bool
            Update the local cache. Defaults to ``True``, but can be
            disabled to reuse the last cached state.
        identity_file: str | pathlib.Path
            The identity file (private key) to use when connecting via
            SSH.
        known_hosts_file: str | pathlib.Path
            The ``known_hosts`` file to pass to SSH for verifying the
            server's host key.
        username: str
            The username to log in as remotely.
        password: str
            The password to use for logging in. Will be ignored when
            ``identity_file`` is passed as well.
        diagram_cache
            An optional place where to find pre-rendered, cached
            diagrams. When a diagram is found in this cache, it will be
            loaded from there instead of being rendered on access. Note
            that diagrams will only be loaded from there, but not be put
            back, i.e. to use it effectively, the cache has to be
            pre-populated.

            This argument accepts the following values:

            - ``None``, in which case all diagrams will be rendered
              internally the first time they are used.
            - A path to a local directory.
            - A URL just like for the ``path`` argument.
            - A dictionary with the arguments to a
              :py:class:`~capellambse.FileHandler`. The dict's ``path``
              key will be analyzed to determine the correct FileHandler
              class.
            - An instance of :py:class:`~capellambse.FileHandler`, which
              will be used directly.

            .. warning:: When using the diagram cache, always make sure
               that the cached diagrams actually match the model version
               that is being used. There is no way to check this
               automatically.

            The file names looked up in the cache built in the format
            ``uuid.ext``, where ``uuid`` is the UUID of the diagram (as
            reported by ``diag_obj.uuid``) and ``ext`` is the render
            format. Example:

            - Diagram ID: ``_7FWu4KrxEeqOgqWuHJrXFA``
            - Render call: ``diag_obj.as_svg`` or ``diag_obj.render("svg")``
            - Cache file name: ``_7FWu4KrxEeqOgqWuHJrXFA.svg``

            *This argument is **not** passed to the file handler.*
        diagram_cache_subdir: str
            A sub-directory prefix to prepend to diagram UUIDs before
            looking them up in the ``diagram_cache``.

            *This argument is **not** passed to the file handler.*
        jupyter_untrusted: bool
            If set to True, restricts or disables some features that are
            unavailable in an untrusted Jupyter environment. Currently
            this only disables the SVG format as rich display option for
            Ipython, which is needed to avoid rendering issues with
            Github's Jupyter notebook viewer.
        fallback_render_aird: bool
            If set to True, enable the internal engine to render
            diagrams that were not found in the pre-rendered cache.
            Defaults to False, which means an exception is raised
            instead. Ignored if no ``diagram_cache`` was specified.

        See Also
        --------
        capellambse.filehandler.FileHandler :
            Abstract super class for file handlers. Contains information
            needed for implementing custom handlers.
        capellambse.filehandler.localfilehandler.LocalFileHandler :
            The file handler responsible for local files and
            directories.
        capellambse.filehandler.gitfilehandler.GitFileHandler :
            The file handler implementing the ``git://`` protocol.
        capellambse.filehandler.http.HTTPFileHandler :
            A simple ``http(s)://`` file handler.
        """
        capellambse.load_model_extensions()

        self._constructed = False
        self._loader = loader.MelodyLoader(path, **kwargs)
        self._fallback_render_aird = fallback_render_aird
        self.jupyter_untrusted = jupyter_untrusted

        try:
            self._pvext = capellambse.pvmt.load_pvmt_from_model(self._loader)
        except ValueError as err:
            LOGGER.warning(
                "Cannot load PVMT extension: %s: %s", type(err).__name__, err
            )
            LOGGER.warning("Property values are not available in this model")
            self._pvext = None

        if diagram_cache:
            if diagram_cache == path:
                self._diagram_cache = self._loader.filehandler
            elif isinstance(diagram_cache, filehandler.FileHandler):
                self._diagram_cache = diagram_cache
            elif isinstance(diagram_cache, cabc.Mapping):
                self._diagram_cache = filehandler.get_filehandler(
                    **diagram_cache
                )
            else:
                self._diagram_cache = filehandler.get_filehandler(
                    diagram_cache
                )
            self._diagram_cache_subdir = pathlib.PurePosixPath(
                diagram_cache_subdir or "."
            )

        self._constructed = True

    @property
    def _element(self) -> etree._Element:
        for tree in self._loader.trees.values():
            if capellambse.helpers.xtype_of(tree.root) in {
                XT_PROJECT,
                XT_LIBRARY,
            }:
                return tree.root
        raise TypeError("No viable root element found")

    @property
    def _model(self) -> MelodyModel:
        return self

    @property
    def parent(self) -> None:
        raise AttributeError("The model root doesn't have a parent object")

    def save(self, **kw: t.Any) -> None:
        """Save the model back to where it was loaded from.

        Parameters
        ----------
        kw
            Additional keyword arguments accepted by the file handler in
            use. Please see the respective documentation for more info.

        See Also
        --------
        capellambse.filehandler.localfilehandler.LocalFileHandler.write_transaction :
            Accepted ``**kw`` when using local directories
        capellambse.filehandler.gitfilehandler.GitFileHandler.write_transaction :
            Accepted ``**kw`` when using ``git://`` and similar URLs

        Notes
        -----
        With a file handler that contacts a remote location (such as the
        :class:`~capellambse.filehandler.gitfilehandler.GitFileHandler`
        with non-local repositories), saving might fail if the local
        state has gone out of sync with the remote state. To avoid this,
        always leave the ``update_cache`` parameter at its default value
        of ``True`` if you intend to save changes.
        """
        self._loader.save(**kw)

    def search(
        self,
        *xtypes: str | type[common.GenericElement],
        below: common.GenericElement | None = None,
    ) -> common.ElementList:
        r"""Search for all elements with any of the given ``xsi:type``\ s.

        If only one xtype is given, the return type will be
        :class:`common.ElementList`, otherwise it will be
        :class:`common.MixedElementList`.

        If no ``xtypes`` are given at all, this method will return an
        exhaustive list of all (semantic) model objects that have an
        ``xsi:type`` set.

        Parameters
        ----------
        xtypes
            The ``xsi:type``\ s to search for, or the classes
            corresponding to them (or a mix of both).
        below
            A model element to constrain the search. If given, only
            those elements will be returned that are (immediate or
            nested) children of this element. This option takes into
            account model fragmentation, but it does not treat link
            elements specially.
        """
        xtypes_: list[str] = []
        for i in xtypes:
            if isinstance(i, type):
                xtypes_.append(common.build_xtype(i))
            elif ":" in i:
                xtypes_.append(i)
            else:
                suffix = ":" + i
                matching_types: list[str] = []
                for l in common.XTYPE_HANDLERS.values():
                    matching_types.extend(t for t in l if t.endswith(suffix))
                if not matching_types:
                    raise ValueError(f"Unknown incomplete type name: {i}")
                xtypes_.extend(matching_types)

        cls = (common.MixedElementList, common.ElementList)[len(xtypes_) == 1]
        trees = {
            k
            for k, v in self._loader.trees.items()
            if v.fragment_type is loader.FragmentType.SEMANTIC
        }
        matches = self._loader.iterall_xt(*xtypes_, trees=trees)
        if below is not None:
            matches = (
                i for i in matches if below._element in i.iterancestors()
            )
        return cls(self, list(matches), common.GenericElement)

    def by_uuid(self, uuid: str) -> common.GenericElement:
        """Search the entire model for an element with the given UUID."""
        return common.GenericElement.from_model(self, self._loader[uuid])

    def find_references(
        self, target: ModelObject | str, /
    ) -> cabc.Iterator[tuple[ModelObject, str, int | None]]:
        """Search the model for references to the given object.

        Parameters
        ----------
        target
            The target object to search for.

        Yields
        ------
        tuple[ModelObject, str, int | None]
            A 3-tuple containing the referencing model object, the
            attribute on that object, and an optional index. If the
            attribute contains a list of objects, the index shows the
            index into the list that was found. Otherwise the index is
            None.
        """
        if isinstance(target, str):
            if not capellambse.helpers.is_uuid_string(target):
                raise ValueError(f"Malformed UUID: {target!r}")
            uuid: str = target
            target = self.by_uuid(target)

        else:
            if target._model is not self:
                raise ValueError(
                    "Cannot find references to objects from different models"
                )
            uuid = getattr(target, "uuid", "")
            if not capellambse.helpers.is_uuid_string(uuid):
                raise ValueError(f"Malformed or missing UUID for {target!r}")

        for elem in self._loader.xpath(
            f"//*[@*[contains(., '#{uuid}')] | */@*[contains(., '#{uuid}')]]",
            roots=[
                i.root
                for i in self._loader.trees.values()
                if i.fragment_type != loader.FragmentType.VISUAL
            ],
        ):
            obj = GenericElement.from_model(self, elem)
            for attr in _reference_attributes(type(obj)):
                if attr.startswith("_"):
                    continue

                try:
                    value = getattr(obj, attr)
                except Exception:
                    continue

                if isinstance(value, GenericElement) and value == target:
                    yield (obj, attr, None)
                elif isinstance(value, ElementList):
                    try:
                        idx = value.index(target)
                    except ValueError:
                        continue
                    yield (obj, attr, idx)

    def update_diagram_cache(
        self,
        capella_cli: str,
        image_format: t.Literal["bmp", "gif", "jpg", "png", "svg"] = "svg",
        *,
        create_index: bool = False,
        force: t.Literal["docker", "exe"] | None = None,
    ) -> None:
        r"""Update the diagram cache if one has been specified.

        If a ``diagram_cache`` has been specified while loading a
        Capella model it will be updated when this function is called.

        The diagram cache will be populated by executing the Capella
        function "Export representations as images" which is normally
        accessible via the context menu of an ``.aird`` node in
        Capella's project explorer. The export of diagrams happens with
        the help of Capella's command line interface (CLI).

        The CLI of Capella must be specified by the caller. It is
        possible to work with a local installation or a Docker image of
        an individual Capella bundle.

        At the moment it is supported to run a Docker image using the
        container system Docker and the ``docker`` executable must be in
        the ``PATH`` environment variable.

        Parameters
        ----------
        capella_cli
            The Capella CLI to use when exporting diagrams from the
            given Capella model. The provided string can come with a
            ``"{VERSION}"`` placeholder. If specified, this placeholder
            will be replaced by the x.y.z formatted version of Capella
            that has been used when the given Capella model was last
            saved. After consideration of the optional placeholder this
            function will first check if the value of ``capella_cli``
            points to a local Capella executable (that can be an
            absolute path or an executable/ symbolic link that has been
            made available via the environment variable ``PATH``). If no
            executable can be found it is expected that ``capella_cli``
            represents a Docker image name for an image that behaves
            like the Capella CLI. For the case of passing a Docker image
            name through ``capella_cli`` this means it is assumed that
            something like the following

            .. code-block:: bash

                docker run --rm -it <capella_cli> -nosplash \
                    -consolelog -app APP -appid APPID

            will work.
            The parameter ``force`` can be set to change the described
            behaviour and force the function to treat the
            ``capella_cli`` as a local executable or a Docker image
            only.
        image_format
            Format of the image file(s) for the exported diagram(s).
            This can be set to any value out of ``"bmp"``, ``"gif"``,
            ``"jpg"``, ``"png"``, or ``"svg"``.
        create_index
            If ``True``, two index files ``index.json`` and
            ``index.html`` will be created. The JSON file consists of a
            list of dictionaries, each representing a diagram in the
            model. The dictionaries come with the keys

            - name: Name of the diagram as it has been set in Capella
            - uuid: The UUID of the diagram
            - success: A boolean stating if a diagram has been exported
              from Capella

            The HTML file shows a numbered list of diagram names which
            are hyperlinked to the diagram image file. Right beside a
            diagram's name one can also see the diagram's UUID in a
            discreet light gray and tiny font size. The HTML index also
            provides some meta data like a timestamp for the
            update of diagrams.
        force
            If the value of ``capella_cli`` is ambiguous and can match
            both a local executable and a Docker image, this parameter
            can be used to bypass the auto-detection and force the
            choice. A value of ``"exe"`` always interprets
            ``capella_cli`` as local executable, ``"docker"`` always
            interprets it as a docker image name. ``None`` (the default)
            enables automatic detection.

        Raises
        ------
        TypeError
            If no ``diagram_cache`` was specified while loading the
            model.
        RuntimeError
            If an error occurs while diagrams are being exported from
            the Capella model.

        Examples
        --------
        **Running a local installation of Capella**

        All the following examples call the method
        :meth:`update_diagram_cache` on a model
        for which a diagram cache has been specified, example:

        >>> import capellambse
        >>> model = capellambse.MelodyModel(
        ...    "/path/to/model.aird",
        ...    diagram_cache="/path/to/diagram_cache",
        ... )

        Passing an executable/ symlink named ``capella`` that is in the
        ``PATH`` environment variable:

        >>> model.update_diagram_cache(
        ...     "capella", "png", True
        ... )

        Passing an absolute path to a local installation of Capella that
        contains the Capella version:

        >>> model.update_diagram_cache(
        ...     "/Applications/Capella_{VERSION}.app/Contents/MacOS/capella"
        ... )

        **Running a Capella CLI container**

        >>> model.update_diagram_cache(
        ...     "capella/cli:{VERSION}-latest"
        ... )
        """
        if not hasattr(self, "_diagram_cache"):
            raise TypeError(
                "Cannot update: No diagram_cache was specified for this model"
            )
        _diagram_cache.export(
            capella_cli,
            self,
            format=image_format,
            index=create_index,
            force=force,
        )

    @classmethod
    def from_model(cls, model: MelodyModel, element: t.Any) -> MelodyModel:
        if element is not model._element:
            raise ValueError("'element' is not the model root")
        return model

    @property
    def info(self) -> loader.ModelInfo:
        return self._loader.get_model_info()

    @property
    def description_badge(self) -> str:
        """Describe model contents distribution with an SVG badge."""
        from capellambse.extensions import metrics

        return metrics.get_summary_badge(self)

    if t.TYPE_CHECKING:

        def __getattr__(self, attr: str) -> t.Any:
            """Account for extension attributes in static type checks."""


def _reference_attributes(objtype: type[ModelObject], /) -> tuple[str, ...]:
    ignored_accessors: tuple[type[common.Accessor], ...] = (
        common.DeepProxyAccessor,
        common.DeprecatedAccessor,
        common.ParentAccessor,
        common.ReferenceSearchingAccessor,
    )

    attrs: list[str] = []
    for i in dir(objtype):
        if i.startswith("_") or i == "parent":
            continue
        acc = getattr(objtype, i, None)
        if isinstance(acc, common.Accessor) and not isinstance(
            acc, ignored_accessors
        ):
            attrs.append(i)
    return tuple(attrs)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import datetime
import operator
import typing as t

import pytest

import capellambse
from capellambse import helpers
from capellambse.extensions import reqif

long_req_text = """\
<p>Test requirement 1 really l o n g text that is&nbsp;way too long to \
display here as that</p>

<p>&lt; &gt; &quot; &#39;</p>

<ul>
\t<li>This&nbsp;is a list</li>
\t<li>an unordered one</li>
</ul>

<ol>
\t<li>Ordered list</li>
\t<li>Ok</li>
</ol>
"""


@pytest.mark.parametrize(
    "expected",
    [
        pytest.param(
            "<Folder 'Folder' (e16f5cc1-3299-43d0-b1a0-82d31a137111)>",
            id="ReqIFName",
        ),
        pytest.param(
            "<Requirement 'TestReq1' (3c2d312c-37c9-41b5-8c32-67578fa52dc3)>",
            id="ReqIFChapterName",
        ),
        pytest.param(
            "<Requirement 'TypedReq2' (0a9a68b1-ba9a-4793-b2cf-4448f0b4b8cc)>",
            id="ReqIFLongName",
        ),
        pytest.param(
            (
                "<CapellaTypesFolder 'Types'"
                " (67bba9cf-953c-4f0b-9986-41991c68d241)>"
            ),
            id="Only ReqIFLongName",
        ),
        pytest.param(
            "<Requirement '' (79291c33-5147-4543-9398-9077d582576d)>",
            id="Empty name",
        ),
        pytest.param(
            (
                "<DateValueAttribute [AttrDef] 2021-07-23T15:00:00+02:00"
                " (b97c09b5-948a-46e8-a656-69d764ddce7d)>"
            ),
            id="Definition ReqIFLongName",
        ),
        pytest.param(
            (
                "<CapellaIncomingRelation 'Controlling the weather' from"
                " <Requirement 'TestReq1'"
                " (3c2d312c-37c9-41b5-8c32-67578fa52dc3)>"
                " to <Entity 'Weather' (4bf0356c-89dd-45e9-b8a6-e0332c026d33)>"
                " (078b2c69-4352-4cf9-9ea5-6573b75e5eec)>"
            ),
            id="IncRelation",
        ),
    ],
)
def test_ReqIFElement_short_repr_(
    model_5_2: capellambse.MelodyModel, expected: str
) -> None:
    r"""Test display of ``ReqIFElement``\ s appearance."""
    matches = helpers.RE_VALID_UUID.findall(expected)
    assert matches is not None
    uuid = matches[-1]
    obj = model_5_2.by_uuid(uuid)

    assert repr(obj).startswith(expected)


def test_extension_was_loaded():
    capellambse.load_model_extensions()

    assert hasattr(capellambse.model.common.GenericElement, "requirements")
    for layer in (
        capellambse.model.layers.oa.OperationalAnalysis,
        capellambse.model.layers.ctx.SystemAnalysis,
        capellambse.model.layers.la.LogicalArchitecture,
        capellambse.model.layers.pa.PhysicalArchitecture,
    ):
        assert hasattr(layer, "requirement_modules")
        assert hasattr(layer, "all_requirements")


def test_path_nesting(model: capellambse.MelodyModel) -> None:
    modules = model.oa.requirement_modules
    assert len(modules) == 2
    assert len(modules[0].folders) == 1
    assert len(modules[0].folders[0].folders) == 1
    assert len(modules[0].folders[0].folders[0].requirements) == 1


class TestRequirementAttributes:
    @pytest.mark.parametrize(
        "attributes",
        [
            pytest.param(
                {
                    "uuid": "e16f5cc1-3299-43d0-b1a0-82d31a137111",
                    "name": "Folder",
                    "long_name": "Folder1",
                    "prefix": "F",
                    "chapter_name": "C",
                    "description": "This is a requirements folder.",
                    "foreign_id": 11,
                    "identifier": "1",
                    "type": reqif.RequirementType,
                    "type.long_name": "ReqType",
                },
                id="Folder",
            ),
            pytest.param(
                {
                    "uuid": "f8e2195d-b5f5-4452-a12b-79233d943d5e",
                    "long_name": "Module",
                    "identifier": "1",
                    "name": "Test Module",
                    "prefix": "T",
                    "description": "This is a test requirement module.",
                    "type": reqif.ModuleType,
                    "type.long_name": "ModuleType",
                },
                id="Module",
            ),
            pytest.param(
                {
                    "uuid": "3c2d312c-37c9-41b5-8c32-67578fa52dc3",
                    "name": "TestReq1",
                    "long_name": "1",
                    "chapter_name": "2",
                    "prefix": "3",
                    "description": "This is a test requirement of kind 1.",
                    "text": long_req_text,
                    "identifier": "REQTYPE-1",
                    "foreign_id": 1,
                },
                id="Requirement",
            ),
            pytest.param(
                {
                    "uuid": "6af6ff84-8957-481f-8684-2405ffa15804",
                    "long_name": "Test Relation",
                    "xtype": "CapellaRequirements:CapellaOutgoingRelation",
                    "description": "This is a relation.",
                    "identifier": "1",
                    "source": reqif.Requirement,
                    "source.name": "TypedReq1",
                    "target": capellambse.model.layers.ctx.SystemFunction,
                    "target.name": "Sysexfunc",
                    "type": reqif.RelationType,
                    "type.long_name": "RelationType",
                },
                id="Relation",
            ),
        ],
    )
    def test_well_defined_generics(
        self,
        model: capellambse.MelodyModel,
        attributes: dict[str, str | int | type],
    ) -> None:
        obj = model.by_uuid(attributes["uuid"])  # type: ignore[arg-type]
        for attr_name, value in attributes.items():
            if isinstance(value, type):
                assert isinstance(operator.attrgetter(attr_name)(obj), value)
            else:
                assert operator.attrgetter(attr_name)(obj) == value

    def test_well_defined_on_RequirementsModules(
        self, model: capellambse.MelodyModel
    ) -> None:
        module = model.by_uuid("f8e2195d-b5f5-4452-a12b-79233d943d5e")
        assert isinstance(module, reqif.CapellaModule)
        attr = module.attributes[0]

        assert len(module.attributes) == 1
        assert isinstance(attr, reqif.EnumerationValueAttribute)
        assert attr.xtype.rsplit(":")[-1] == "EnumerationValueAttribute"
        assert isinstance(
            attr.definition, reqif.AttributeDefinitionEnumeration
        )
        assert attr.definition.long_name == "AttrDefEnum"
        assert attr.values[0].long_name == "enum_val2"
        assert attr.values[0] == attr.value

    def test_well_defined_on_Requirements(
        self, model: capellambse.MelodyModel
    ) -> None:
        req = model.by_uuid("3c2d312c-37c9-41b5-8c32-67578fa52dc3")
        req2 = model.by_uuid("0a9a68b1-ba9a-4793-b2cf-4448f0b4b8cc")
        assert isinstance(req, reqif.Requirement)
        assert isinstance(req2, reqif.Requirement)

        bool_attr = req.attributes[0]
        undefined_attr = req.attributes[-1]

        assert len(req.attributes) == 6
        assert undefined_attr.value is False
        assert isinstance(bool_attr.value, bool)
        for attr, typ in zip(req.attributes[2:-1], [int, float, str]):
            assert isinstance(attr.value, typ)

        actual_values = [req.long_name for req in req2.attributes[0].values]
        assert actual_values == ["enum_val1", "enum_val2"]

    @pytest.mark.parametrize(
        ["uuid", "description"],
        [
            (
                "9c692405-b8aa-4caa-b988-51d27db5cd1b",
                "BooleanValueAttribute [AttrDef] True",
            ),
            (
                "b97c09b5-948a-46e8-a656-69d764ddce7d",
                "DateValueAttribute [AttrDef] 2021-07-23T15:00:00+02:00",
            ),
            (
                "85dfd42c-7f6e-4236-a181-bdd784040431",
                "IntegerValueAttribute [AttrDef] 10",
            ),
            (
                "d2231d14-854d-4625-b48b-6cf1c2554367",
                "RealValueAttribute [AttrDef] 10.5",
            ),
            (
                "ee8a69ef-61b9-4db9-9a0f-628e5d4704e1",
                "StringValueAttribute [AttrDef] 'Test'",
            ),
            (
                "dcb8614e-2d1c-4cb3-aa0c-667a297e7489",
                "BooleanValueAttribute [Boolean Value Attribute] False",
            ),
            (
                "148bdf2f-6dc2-4a83-833b-596886ce5b07",
                (
                    "EnumerationValueAttribute [MultiEnum]"
                    " ['enum_val1', 'enum_val2']"
                ),
            ),
        ],
    )
    def test_repr_shows_value(
        self, model: capellambse.MelodyModel, uuid: str, description: str
    ) -> None:
        attribute = model.by_uuid(uuid)

        assert repr(attribute) == f"<{description} ({uuid})>"


class TestRequirementRelations:
    def test_well_defined_source_target_and_type(
        self, model: capellambse.MelodyModel
    ) -> None:
        rel = model.by_uuid("078b2c69-4352-4cf9-9ea5-6573b75e5eec")
        source = model.by_uuid("3c2d312c-37c9-41b5-8c32-67578fa52dc3")
        target = model.by_uuid("4bf0356c-89dd-45e9-b8a6-e0332c026d33")
        assert isinstance(rel, reqif.AbstractRequirementsRelation)

        assert rel.source == source
        assert rel.target == target
        assert isinstance(rel.type, reqif.RelationType)
        assert rel.type.long_name == "RelationType"

    def test_well_defined_on_Requirements(
        self, model: capellambse.MelodyModel
    ) -> None:
        req = model.by_uuid("3c2d312c-37c9-41b5-8c32-67578fa52dc3")
        assert isinstance(req, reqif.Requirement)

        assert len(req.relations) == 4

    def test_well_defined_on_GenericElements(
        self, model: capellambse.MelodyModel
    ) -> None:
        ge = model.by_uuid("00e7b925-cf4c-4cb0-929e-5409a1cd872b")

        assert isinstance(ge.requirements, reqif.RelationsList)
        assert len(ge.requirements) == 3

    def test_filtering_by_relation_type(self, model: capellambse.MelodyModel):
        ge = model.by_uuid("00e7b925-cf4c-4cb0-929e-5409a1cd872b")
        rel_type = model.by_uuid("f1aceb81-5f70-4469-a127-94830eb9be04")

        assert len(ge.requirements.by_relation_type(rel_type.name)) == 1

    @pytest.mark.parametrize(
        ("obj_uuid", "target_uuids"),
        [
            (
                "3c2d312c-37c9-41b5-8c32-67578fa52dc3",
                {
                    "078b2c69-4352-4cf9-9ea5-6573b75e5eec",
                    "24c824ef-b187-4725-a051-a68707e82d70",
                    "57033242-3766-4961-8091-ce3d9326ed67",
                    "7de4c1a5-e106-4171-902a-502b816b60b0",
                },
            ),
            (
                "79291c33-5147-4543-9398-9077d582576d",
                {"41e1b786-a6a1-46cb-9b9c-b302d9278c1c"},
            ),
        ],
    )
    def test_RequirementRelations(
        self, model: capellambse.MelodyModel, obj_uuid, target_uuids
    ):
        obj = model.by_uuid(obj_uuid)
        assert isinstance(obj, reqif.Requirement)

        relations = obj.relations

        assert isinstance(relations, capellambse.model.ElementList)
        assert all(
            isinstance(i, reqif.AbstractRequirementsRelation)
            for i in relations
        )
        assert {i.uuid for i in relations} == target_uuids


class TestReqIFAccess:
    def test_RequirementsModule_attributes(
        self, model: capellambse.MelodyModel
    ):
        mod = model.by_uuid("f8e2195d-b5f5-4452-a12b-79233d943d5e")
        assert isinstance(mod, reqif.CapellaModule)

        assert len(mod.folders) == 1
        assert len(mod.requirements) == 1
        assert mod.type.long_name == "ModuleType"
        for attr, expected in {
            "identifier": "1",
            "long_name": "Module",
            "name": "Test Module",
            "prefix": "T",
            "description": "This is a test requirement module.",
        }.items():
            assert getattr(mod, attr) == expected

    def test_RequirementFolder_attributes(
        self, model: capellambse.MelodyModel
    ):
        folder = model.by_uuid("e16f5cc1-3299-43d0-b1a0-82d31a137111")
        assert isinstance(folder, reqif.Folder)

        assert len(folder.folders) == 1
        assert len(folder.requirements) == 2
        assert folder.type.long_name == "ReqType"
        for attr, expected in {
            "identifier": "1",
            "long_name": "Folder1",
            "name": "Folder",
            "prefix": "F",
            "chapter_name": "C",
            "foreign_id": 11,
            "text": "This is a folder.",
            "description": "This is a requirements folder.",
        }.items():
            assert getattr(folder, attr) == expected

    def test_Requirement_attributes(self, model: capellambse.MelodyModel):
        req = model.by_uuid("3c2d312c-37c9-41b5-8c32-67578fa52dc3")
        assert isinstance(req, reqif.Requirement)
        assert req.type.long_name == "ReqType"

        for attr, expected in {
            "chapter_name": "2",
            "description": "This is a test requirement of kind 1.",
            "foreign_id": 1,
            "identifier": "REQTYPE-1",
            "long_name": "1",
            "name": "TestReq1",
            "prefix": "3",
            "text": long_req_text,
        }.items():
            assert getattr(req, attr) == expected

    def test_Relations(self, model: capellambse.MelodyModel):
        req_with_relations = model.by_uuid(
            "3c2d312c-37c9-41b5-8c32-67578fa52dc3"
        )
        assert isinstance(req_with_relations, reqif.Requirement)

        relations = req_with_relations.relations
        assert len(relations) == 4

    def test_Requirement_without_Relations(
        self, model: capellambse.MelodyModel
    ):
        req_without_relations = model.by_uuid(
            "0a9a68b1-ba9a-4793-b2cf-4448f0b4b8cc"
        )
        assert isinstance(req_without_relations, reqif.Requirement)
        assert len(req_without_relations.relations) == 0

    def test_outgoing_and_internal_Relations(
        self, model: capellambse.MelodyModel
    ):
        req_with_oir = model.by_uuid("85d41db2-9e17-438b-95cf-49342452ddf3")
        assert isinstance(req_with_oir, reqif.Requirement)
        assert len(req_with_oir.relations) == 2

    def test_RequirementTypes_AttributeDefinitions(
        self, model: capellambse.MelodyModel
    ):
        reqtype = model.by_uuid("db47fca9-ddb6-4397-8d4b-e397e53d277e")
        attr_def = model.by_uuid("682bd51d-5451-4930-a97e-8bfca6c3a127")
        enum_def = model.by_uuid("c316ab07-c5c3-4866-a896-92e34733055c")

        assert attr_def in reqtype.attribute_definitions
        assert enum_def in reqtype.attribute_definitions


class TestReqIFModification:
    def test_created_Requirements_can_be_found_in_the_model(
        self, model: capellambse.MelodyModel
    ):
        mod = model.oa.requirement_modules[0]
        new_req = mod.requirements.create("Requirement")

        assert model.by_uuid(new_req.uuid) == new_req
        assert new_req in mod.requirements

    def test_deleted_Requirements_vanish_from_model(
        self, model: capellambse.MelodyModel
    ):
        mod = model.oa.requirement_modules[0]
        old_req = mod.requirements[0]

        del mod.requirements[0]

        assert old_req not in mod.requirements
        with pytest.raises(KeyError):
            model.by_uuid(old_req.uuid)

    @pytest.mark.parametrize(
        "relcls",
        [
            pytest.param("CapellaIncomingRelation", id="IncRelation"),
            pytest.param("CapellaOutgoingRelation", id="OutRelation"),
            pytest.param("InternalRelation", id="IntRelation"),
        ],
    )
    def test_creating_Requirements_raises_TypeError(
        self, model: capellambse.MelodyModel, relcls: str
    ):
        req = model.by_uuid("3c2d312c-37c9-41b5-8c32-67578fa52dc3")
        assert isinstance(req, reqif.Requirement)

        with pytest.raises(TypeError):
            req.relations.create(relcls)
        with pytest.raises(TypeError):
            req.relations.create(
                relcls, target="e16f5cc1-3299-43d0-b1a0-82d31a137111"
            )
        with pytest.raises(TypeError):
            req.relations.create(relcls, type="RelationType")
        with pytest.raises(TypeError):
            req.relations.create(relcls, target=req.attributes[0].definition)

    def test_created_Requirements_are_found_from_both_sides(
        self, model: capellambse.MelodyModel
    ):
        req = model.by_uuid("3c2d312c-37c9-41b5-8c32-67578fa52dc3")
        target = model.by_uuid("79291c33-5147-4543-9398-9077d582576d")
        assert isinstance(req, reqif.Requirement)
        assert isinstance(target, reqif.Requirement)

        reltype = model.by_uuid("f1aceb81-5f70-4469-a127-94830eb9be04")
        new_rel = req.relations.create(target=target, type=reltype)

        assert isinstance(req.relations, capellambse.model.ElementList)
        assert new_rel in req.relations
        assert isinstance(target.relations, capellambse.model.ElementList)
        assert new_rel in target.relations

    TEST_DATETIME = datetime.datetime(1987, 7, 27)
    TEST_TZ_DELTA = TEST_DATETIME.astimezone().utcoffset()
    assert TEST_TZ_DELTA is not None
    TEST_TZ_HRS, TEST_TZ_SECS = divmod(TEST_TZ_DELTA.seconds, 3600)
    TEST_TZ_MINS, TEST_TZ_SECS = divmod(TEST_TZ_SECS, 60)
    TEST_TZ_OFFSET = f"{TEST_TZ_HRS:+03d}{TEST_TZ_MINS:02d}"

    @pytest.mark.parametrize(
        "type_hint,default_value",
        [
            pytest.param("Int", 0),
            pytest.param("Str", ""),
            pytest.param("Float", 0.0),
            pytest.param("Date", None),
            pytest.param("Bool", False),
        ],
    )
    def test_create_ValueAttributes_with_default_values(
        self,
        model: capellambse.MelodyModel,
        type_hint: str,
        default_value: t.Any,
    ):
        req = model.by_uuid("79291c33-5147-4543-9398-9077d582576d")

        assert isinstance(req, reqif.Requirement)
        assert not req.attributes

        definition = model.by_uuid("682bd51d-5451-4930-a97e-8bfca6c3a127")
        attr = req.attributes.create(type_hint, definition=definition)

        assert req.attributes == [attr]
        assert attr.definition == definition
        assert attr.value == default_value
        assert attr._element.get("value") is None
        assert isinstance(attr, reqif.AbstractRequirementsAttribute)

    @pytest.mark.parametrize(
        "type_hint,value,xml",
        [
            pytest.param("Int", 0, None),
            pytest.param("Int", 1, "1"),
            pytest.param("Str", "", None),
            pytest.param("Str", "test", "test"),
            pytest.param("Float", 0.0, None),
            pytest.param("Float", 1.0, "1.0"),
            pytest.param("Date", None, None),
            pytest.param(
                "Date",
                TEST_DATETIME,
                f"1987-07-27T00:00:00.000{TEST_TZ_OFFSET}",
            ),
            pytest.param(
                "Date",
                datetime.datetime(1987, 7, 27, tzinfo=datetime.timezone.utc),
                "1987-07-27T00:00:00.000+0000",
            ),
            pytest.param("Bool", False, None),
            pytest.param("Bool", True, "true"),
        ],
    )
    def test_create_ValueAttributes_with_non_default_values(
        self,
        model: capellambse.MelodyModel,
        type_hint: str,
        value: t.Any,
        xml: str | None,
    ):
        req = model.by_uuid("79291c33-5147-4543-9398-9077d582576d")

        assert isinstance(req, reqif.Requirement)
        assert not req.attributes

        value_attr = req.attributes.create(type_hint, value=value)
        if isinstance(value, datetime.datetime):
            value = value.astimezone()

        assert req.attributes == [value_attr]
        assert value_attr.value == value
        assert value_attr._element.get("value") == xml

    def test_create_ValueAttribute_on_Requirements_without_definition(
        self, model: capellambse.MelodyModel
    ):
        req = model.by_uuid("79291c33-5147-4543-9398-9077d582576d")
        assert isinstance(req, reqif.Requirement)
        assert not req.attributes

        attr = req.attributes.create("Int")

        assert len(req.attributes) == 1
        assert req.attributes[0] == attr
        assert attr.definition is None
        assert isinstance(attr, reqif.AbstractRequirementsAttribute)

    @pytest.mark.parametrize(
        "type_hint,expected_type",
        [
            pytest.param("Int", "Integer"),
            pytest.param("Integer", "Integer"),
            pytest.param("Integervalueattribute", "Integer"),
            pytest.param("Str", "String"),
            pytest.param("String", "String"),
            pytest.param("Stringvalueattribute", "String"),
            pytest.param("Float", "Real"),
            pytest.param("Real", "Real"),
            pytest.param("Realvalueattribute", "Real"),
            pytest.param("Date", "Date"),
            pytest.param("Datevalueattribute", "Date"),
            pytest.param("Bool", "Boolean"),
            pytest.param("Boolean", "Boolean"),
            pytest.param("Booleanvalueattribute", "Boolean"),
        ],
    )
    def test_Requirements_ValueAttribute_default_reprs(
        self,
        model: capellambse.MelodyModel,
        type_hint: str,
        expected_type: str,
    ):
        req = model.by_uuid("79291c33-5147-4543-9398-9077d582576d")

        assert isinstance(req, reqif.Requirement)

        attr = req.attributes.create(type_hint)

        assert f"[{expected_type} Value Attribute]" in repr(attr)

    def test_create_EnumValueAttribute_on_Requirements(
        self, model: capellambse.MelodyModel
    ):
        req = model.by_uuid("79291c33-5147-4543-9398-9077d582576d")
        definition = model.by_uuid("c316ab07-c5c3-4866-a896-92e34733055c")
        assert isinstance(req, reqif.Requirement)

        assert not req.attributes
        attr = req.attributes.create("Enum", definition=definition)

        assert len(req.attributes) == 1
        assert req.attributes[0] == attr
        assert attr.definition == definition
        assert isinstance(attr, reqif.EnumerationValueAttribute)

    def test_create_EnumValueAttribute_with_passing_values(
        self, model: capellambse.MelodyModel
    ):
        req = model.oa.all_requirements[0]
        dtdef = model.by_uuid("637caf95-3229-4607-99a0-7d7b990bc97f")
        values = [
            model.by_uuid("efd6e108-3461-43c6-ad86-24168339ed3c"),
            model.by_uuid("3c2390a4-ce9c-472c-9982-d0b825931978"),
        ]

        attr = req.attributes.create("Enum", values=values)

        assert attr in req.attributes
        assert attr.values == dtdef.values

    def test_create_ValueAttribute_with_wrong_type_hint_raises_ValueError(
        self, model: capellambse.MelodyModel
    ):
        req = model.by_uuid("79291c33-5147-4543-9398-9077d582576d")
        assert isinstance(req, reqif.Requirement)

        with pytest.raises(ValueError):
            req.attributes.create("Gibberish")

    @pytest.mark.parametrize(
        "value",
        [
            pytest.param(True, id="Boolean Attribute"),
            pytest.param(False, id="Boolean Attribute"),
            pytest.param(1, id="Integer Attribute"),
            pytest.param(
                datetime.datetime(
                    2021, 7, 23, 13, 0, 0, tzinfo=datetime.timezone.utc
                ),
                id="DateValue Attribute",
            ),
            pytest.param(1.9, id="Float/Real Attribute"),
            pytest.param("Test1", id="String Attribute"),
        ],
    )
    def test_setting_ValueAttributes_on_Requirement(
        self, model: capellambse.MelodyModel, value: t.Any
    ):
        req = model.by_uuid("3c2d312c-37c9-41b5-8c32-67578fa52dc3")
        definition = model.by_uuid("682bd51d-5451-4930-a97e-8bfca6c3a127")
        assert isinstance(req, reqif.Requirement)

        attributes = req.attributes.by_definition(definition)
        attr = next(
            attr for attr in attributes if type(attr.value) is type(value)
        )
        attr.value = value

        assert attr.value == value

    @pytest.mark.parametrize(
        "default_value",
        [
            pytest.param(False, id="Boolean Attribute"),
            pytest.param(0, id="Integer Attribute"),
            pytest.param(None, id="DateValue Attribute"),
            pytest.param(0.0, id="Float/Real Attribute"),
            pytest.param("", id="String Attribute"),
        ],
    )
    def test_setting_default_value_removes_value_on_xml_element(
        self,
        model: capellambse.MelodyModel,
        default_value: t.Any,
    ):
        req = model.by_uuid("3c2d312c-37c9-41b5-8c32-67578fa52dc3")
        definition = model.by_uuid("682bd51d-5451-4930-a97e-8bfca6c3a127")
        assert isinstance(req, reqif.Requirement)

        attributes = req.attributes.by_definition(definition)
        try:
            attr = next(
                attr
                for attr in attributes
                if type(attr.value) is type(default_value)
            )
        except StopIteration:
            assert default_value is None
            attr = attributes[1]

        attr.value = default_value

        assert attr.value == default_value
        assert "value" not in attr._element.attrib

    @pytest.mark.parametrize(
        "uuid,value",
        [
            pytest.param("b97c09b5-948a-46e8-a656-69d764ddce7d", 1, id="Int"),
        ],
    )
    def test_setting_ValueAttribute_with_wrong_type_fails_with_TypeError(
        self, model: capellambse.MelodyModel, uuid: str, value: t.Any
    ):
        attr = model.by_uuid(uuid)
        assert isinstance(attr, reqif.AbstractRequirementsAttribute)

        with pytest.raises(TypeError):
            attr.value = value

    def test_setting_values_on_EnumValueAttribute(
        self, model: capellambse.MelodyModel
    ):
        evattr = model.by_uuid("a44b41f3-598c-4250-bca6-1329647e7a02")
        values = [
            model.by_uuid("efd6e108-3461-43c6-ad86-24168339ed3c"),
            model.by_uuid("3c2390a4-ce9c-472c-9982-d0b825931978"),
        ]
        assert values[1] not in evattr.values, "Precondition failed"

        evattr.values = values

        assert evattr.values == values

    def test_create_RequirementType_AttributeDefinition_creation(
        self, model: capellambse.MelodyModel
    ):
        reqtype = model.by_uuid("db47fca9-ddb6-4397-8d4b-e397e53d277e")
        definitions = reqtype.attribute_definitions

        attr_def = reqtype.attribute_definitions.create(
            "AttributeDefinition", long_name="First"
        )
        enum_def = reqtype.attribute_definitions.create(
            "AttributeDefinitionEnumeration", long_name="Second"
        )

        assert len(definitions) + 2 == len(reqtype.attribute_definitions)
        assert attr_def in reqtype.attribute_definitions
        assert enum_def in reqtype.attribute_definitions

    def test_create_EnumDataTypeDefinition_setting_EnumValues(
        self, model: capellambse.MelodyModel
    ):
        reqtypesfolder = model.by_uuid("67bba9cf-953c-4f0b-9986-41991c68d241")
        dt_definitions = reqtypesfolder.data_type_definitions

        edt_def = reqtypesfolder.data_type_definitions.create(
            "EnumerationDataTypeDefinition",
            long_name="Enum",
        )
        edt_def.values.create(long_name="val")
        edt_def.values.create(long_name="val1")

        assert len(dt_definitions) + 1 == len(
            reqtypesfolder.data_type_definitions
        )
        assert edt_def in reqtypesfolder.data_type_definitions
        assert set(edt_def.values.by_long_name) == {"val", "val1"}

    def test_create_EnumDataTypeDefinition_creating_EnumValues(
        self,
        model: capellambse.MelodyModel,
    ):
        reqtypesfolder = model.by_uuid("67bba9cf-953c-4f0b-9986-41991c68d241")
        dt_definitions = reqtypesfolder.data_type_definitions

        edt_def = reqtypesfolder.data_type_definitions.create(
            "EnumerationDataTypeDefinition",
            long_name="Enum",
            values=["val", "val1"],
        )

        assert len(dt_definitions) + 1 == len(
            reqtypesfolder.data_type_definitions
        )
        assert edt_def in reqtypesfolder.data_type_definitions
        assert set(edt_def.values.by_long_name) == {"val", "val1"}


class TestRequirementsFiltering:
    uuids = frozenset(
        {
            "3c2d312c-37c9-41b5-8c32-67578fa52dc3",
            "0a9a68b1-ba9a-4793-b2cf-4448f0b4b8cc",
            "79291c33-5147-4543-9398-9077d582576d",
            "85d41db2-9e17-438b-95cf-49342452ddf3",
            "1092f69a-5f3a-4fe6-a8fd-b2dffde90650",
            "32d66740-e428-4600-be68-8410d9d568cc",
        }
    )
    reqtype_uuid = "db47fca9-ddb6-4397-8d4b-e397e53d277e"

    def test_filtering_by_type_name(self, model: capellambse.MelodyModel):
        requirements = model.search(reqif.Requirement)
        rt = model.by_uuid(self.reqtype_uuid)
        assert isinstance(rt, reqif.RequirementType)

        rtype_reqs = requirements.by_type(rt.long_name)

        uuids = {r.uuid for r in rtype_reqs}
        assert uuids & self.uuids == self.uuids

    def test_filtering_by_type_names(self, model: capellambse.MelodyModel):
        requirements = model.search(reqif.Requirement)
        rt1 = model.by_uuid(self.reqtype_uuid)
        rt2 = model.by_uuid("00195554-8fae-4687-86ca-77c93330893d")
        assert isinstance(rt1, reqif.RequirementType)
        assert isinstance(rt2, reqif.RequirementType)

        rtype_reqs = requirements.by_type(rt1.long_name, rt2.long_name)
        expected_uuids = self.uuids | {"db4d4c74-b913-4d9a-8905-7d93d3360560"}

        uuids = {r.uuid for r in rtype_reqs}
        assert uuids & expected_uuids == expected_uuids

    def test_filtering_by_RequirementType(
        self, model: capellambse.MelodyModel
    ):
        requirements = model.search(reqif.Requirement)
        rt = model.by_uuid(self.reqtype_uuid)

        uuids = {r.uuid for r in requirements.by_type(rt)}
        assert uuids & self.uuids == self.uuids

    def test_filtering_by_types(self, model: capellambse.MelodyModel):
        requirements = model.search(reqif.Requirement)
        rt1 = model.by_uuid(self.reqtype_uuid)
        rt2 = model.by_uuid("00195554-8fae-4687-86ca-77c93330893d")
        expected_uuids = self.uuids | {"db4d4c74-b913-4d9a-8905-7d93d3360560"}

        uuids = {r.uuid for r in requirements.by_type(rt1, rt2)}
        assert uuids & expected_uuids == expected_uuids

    def test_filtering_by_name_and_type(self, model: capellambse.MelodyModel):
        requirements = model.search(reqif.Requirement)
        rt1 = model.by_uuid(self.reqtype_uuid)
        rt2 = model.by_uuid("00195554-8fae-4687-86ca-77c93330893d")
        assert isinstance(rt1, reqif.RequirementType)
        assert isinstance(rt2, reqif.RequirementType)

        expected_uuids = self.uuids | {"db4d4c74-b913-4d9a-8905-7d93d3360560"}

        uuids = {r.uuid for r in requirements.by_type(rt1, rt2.long_name)}
        assert uuids & expected_uuids == expected_uuids

    def test_filtering_by_type_None_returns_requirements_with_undefined_type(
        self, model: capellambse.MelodyModel
    ):
        requirements = model.search(reqif.Requirement)
        undefined_type_reqs = requirements.by_type(None)
        req = undefined_type_reqs[0]

        assert len(undefined_type_reqs) == 1
        assert req.long_name == "Undefined type"
        assert req.uuid == "7e8d0edf-67f0-48c5-82f6-ec9cdb809eee"

    @pytest.mark.parametrize(
        ("obj_uuid", "relation_class", "target_uuids"),
        [
            (
                "00e7b925-cf4c-4cb0-929e-5409a1cd872b",
                "outgoing",
                {"85d41db2-9e17-438b-95cf-49342452ddf3"},
            ),
            (
                "00e7b925-cf4c-4cb0-929e-5409a1cd872b",
                "incoming",
                {
                    "3c2d312c-37c9-41b5-8c32-67578fa52dc3",
                    "79291c33-5147-4543-9398-9077d582576d",
                },
            ),
            ("00e7b925-cf4c-4cb0-929e-5409a1cd872b", "internal", set()),
            (
                "3c2d312c-37c9-41b5-8c32-67578fa52dc3",
                "outgoing",
                {"0d2edb8f-fa34-4e73-89ec-fb9a63001440"},
            ),
            (
                "3c2d312c-37c9-41b5-8c32-67578fa52dc3",
                "incoming",
                {
                    "4bf0356c-89dd-45e9-b8a6-e0332c026d33",
                    "00e7b925-cf4c-4cb0-929e-5409a1cd872b",
                },
            ),
            (
                "3c2d312c-37c9-41b5-8c32-67578fa52dc3",
                "internal",
                {"85d41db2-9e17-438b-95cf-49342452ddf3"},
            ),
        ],
    )
    def test_filtering_by_relation_class(
        self,
        model: capellambse.MelodyModel,
        obj_uuid: str,
        relation_class: str,
        target_uuids: t.Sequence[str],
    ):
        obj = model.by_uuid(obj_uuid)

        if isinstance(obj, reqif.Requirement):
            related = obj.related
        else:
            related = obj.requirements
        filtered_related = related.by_relation_class(relation_class)

        assert isinstance(filtered_related, capellambse.model.ElementList)
        assert {i.uuid for i in filtered_related} == target_uuids

    def test_attributes_filtering_by_definition_long_name(
        self, model: capellambse.MelodyModel
    ):
        req = model.by_uuid("85d41db2-9e17-438b-95cf-49342452ddf3")
        ex_definition = model.by_uuid("c316ab07-c5c3-4866-a896-92e34733055c")
        def_name = ex_definition.long_name

        attributes = req.attributes.by_definition.long_name(def_name)
        attr = req.attributes.by_definition.long_name(def_name, single=True)

        assert def_name in req.attributes.by_definition.long_name
        assert [attr.definition for attr in attributes] == [ex_definition]
        assert attr.definition == ex_definition

    def test_RelationsLists_slicing(self, model: capellambse.MelodyModel):
        req = model.by_uuid("3c2d312c-37c9-41b5-8c32-67578fa52dc3")
        assert isinstance(req, reqif.Requirement)
        related_objs = req.related
        assert isinstance(related_objs, capellambse.model.ElementList)

        sliced = related_objs[:]

        assert isinstance(sliced, capellambse.model.ElementList)
        assert not isinstance(
            sliced, capellambse.model.common.ElementListCouplingMixin
        )
        assert len(sliced) == 4
        assert related_objs is not sliced

    def test_Requirement_is_hashable(self, model: capellambse.MelodyModel):
        req = model.by_uuid("3c2d312c-37c9-41b5-8c32-67578fa52dc3")

        assert isinstance(req, reqif.Requirement)
        assert isinstance(req, t.Hashable)
        assert hash(req)

    def test_RequirementType_is_hashable(self, model: capellambse.MelodyModel):
        req_type = model.by_uuid(self.reqtype_uuid)

        assert isinstance(req_type, reqif.RequirementType)
        assert isinstance(req_type, t.Hashable)
        assert hash(req_type)

    def test_EnumValue_is_hashable(self, model: capellambse.MelodyModel):
        enum_value = model.by_uuid("efd6e108-3461-43c6-ad86-24168339ed3c")

        assert isinstance(enum_value, reqif.EnumValue)
        assert isinstance(enum_value, t.Hashable)
        assert hash(enum_value)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=abstract-method, useless-suppression
# For some reason, pylint in Github CI didn't get the memo that these aren't
# actually abstract methods. Other pylint installations seem to agree that
# implementing these methods isn't necessary. So we just ignore the warning
# about that here.
# TODO Revisit this decision some time in the future

from __future__ import annotations

import collections.abc as cabc
import functools
import hashlib
import io
import itertools
import logging
import os
import pathlib
import re
import shlex
import shutil
import subprocess
import tempfile
import textwrap
import typing as t
import urllib.parse
import weakref

import capellambse.helpers
from capellambse.loader import modelinfo

from . import FileHandler, TransactionClosedError

LOGGER = logging.getLogger(__name__)

_git_object_name = re.compile("(^|/)([0-9a-fA-F]{4,}|(.+_)?HEAD)$")


class _TreeEntry(t.NamedTuple):
    mode: str
    type: str
    object: str
    file: str

    @classmethod
    def fromstring(cls, entry: str) -> _TreeEntry:
        info, file = entry.split("\t", maxsplit=1)
        mode, type, object = info.split(" ")
        return cls(mode, type, object, file)

    def tostring(self) -> str:
        return f"{self.mode} {self.type} {self.object}\t{self.file}"


class _ProcessWriter(t.BinaryIO):
    def __init__(
        self,
        command: cabc.Sequence[t.Any],
        *,
        cwd: pathlib.Path,
        env: dict[str, str],
    ) -> None:
        # pylint: disable=consider-using-with
        LOGGER.debug("Spawning process to stream to: %r", command)
        self.process = subprocess.Popen(
            command,
            cwd=cwd,
            env=env,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )

        assert self.process.stdin is not None
        self.write = self.process.stdin.write  # type: ignore[assignment]

    def write(self, s: bytes) -> int:
        return len(s)  # stub

    def close(self) -> None:
        assert self.process.stdin is not None
        assert self.process.stdout is not None
        if not self.process.stdin.closed:
            self.process.stdin.close()
        if not self.process.stdout.closed:
            self.process.stdout.close()
        self.process.wait()
        assert self.process.returncode is not None
        if self.process.returncode != 0:
            raise RuntimeError(
                f"Subprocess returned error {self.process.returncode}"
            )

    @property
    def closed(self) -> bool:
        assert self.process.stdin is not None
        return self.process.stdin.closed

    def __enter__(self) -> _ProcessWriter:
        return self

    def __exit__(self, *args: t.Any) -> None:
        self.close()


class _WritableIndexFile(_ProcessWriter):
    def __init__(
        self,
        cb: cabc.Callable[[str], None],
        cwd: pathlib.Path,
        env: dict[str, str],
        filename: pathlib.PurePosixPath,
    ) -> None:
        super().__init__(
            ["git", "hash-object", "--path", filename, "--stdin", "-w"],
            cwd=cwd,
            env=env,
        )
        self.callback = cb

    def close(self) -> None:
        assert self.process.stdin is not None
        assert self.process.stdout is not None

        hash, _ = self.process.communicate()
        assert hash
        super().close()
        self.callback(hash.decode("ascii").strip())


class _WritableLFSFile(_ProcessWriter):
    def __init__(
        self,
        cb: cabc.Callable[[str], None],
        cwd: pathlib.Path,
        env: dict[str, str],
        filename: pathlib.PurePosixPath,
    ) -> None:
        super().__init__(
            ["git", "lfs", "clean", "--", filename], cwd=cwd, env=env
        )
        self.__indexfile = _WritableIndexFile(  # type: ignore[abstract]
            cb, cwd, env, filename
        )

    def close(self) -> None:
        assert self.process.stdin is not None
        assert self.process.stdout is not None

        try:
            pointer, _ = self.process.communicate()
            super().close()

            assert pointer
            self.__indexfile.write(pointer)
        finally:
            self.__indexfile.close()


class _GitTransaction:
    __old_sha: str
    __unclosed_error = textwrap.dedent(
        """\
        %d file(s) is/are still opened.

        =======================================================================

        You still have open files when committing a transaction.

        Files that are still open CANNOT be committed and are LOST FOREVER!

        Fix your code to ``close()`` all opened files BEFORE control flows out
        of the ``with:`` block opened by ``write_transaction()``.

        Again: YOU HAVE JUST LOST DATA.

        =======================================================================
        """
    )

    def __init__(
        self,
        outer_transactor: cabc.Callable[..., t.ContextManager],
        filehandler: GitFileHandler,
        /,
        *,
        dry_run: bool = False,
        author_name: str | None = None,
        author_email: str | None = None,
        commit_msg: str = "Changes made with python-capellambse",
        remote_branch: str | None = None,
        push: bool = True,
        push_options: cabc.Sequence[str] = (),
        **kw: t.Any,
    ) -> None:
        """Create a transaction that records all changes as a new commit.

        Parameters
        ----------
        author_name
            The name of the commit author.
        author_email
            The e-mail address of the commit author.
        commit_msg
            The commit message.
        dry_run
            If True, stop before updating the ``revision`` pointer. The
            commit will be created, but will not be part of any branch
            or tag.
        remote_branch
            An alternative branch name to push to on the remote, instead
            of pushing back to the same branch.  This is required if
            ``push`` is ``True`` and the ``revision`` that was passed to
            the constructor does not refer to a branch (or looks like a
            git object).

            Note: For convenience, ``refs/heads/`` will be prepended
            automatically to this name if it isn't already present. This
            also means that it is not possible to create tags or other
            types of refs; passing in something like ``refs/tags/v2.4``
            would result in the full ref name
            ``refs/heads/refs/tags/v2.4``.
        push
            Set to ``False`` to inhibit pushing the changes back.
        push_options
            Additional git push options.  See ``--push-option`` in
            ``git-push(1)``. Ignored if ``push`` is ``False``.

        Raises
        ------
        ValueError
            - If a commit hash was used during loading, and no
              ``remote_branch`` was given
            - If the given ``remote_branch`` (or the final part of the
              originally given revision) looks like a git object
        """
        self.__updates: dict[pathlib.PurePosixPath, str] = {}
        self.__outer_context = outer_transactor(**kw)
        self.__handler = filehandler
        self.__dry_run = dry_run
        self.__commit_msg = commit_msg
        self.__push = push
        self.__push_opts = [f"--push-option={i}" for i in push_options]

        self.__gitenv: dict[str, str] = {}
        if author_name:
            self.__gitenv["GIT_AUTHOR_NAME"] = author_name
        if author_email:
            self.__gitenv["GIT_AUTHOR_EMAIL"] = author_email

        targetref = remote_branch or filehandler.revision

        if targetref == "HEAD":
            targetref = self.__resolve_head() or targetref

        if _git_object_name.search(targetref):
            raise ValueError(
                "Target ref looks like a git object, use a different"
                f" remote_branch: {targetref}"
            )
        if not targetref.startswith("refs/heads/"):
            targetref = "refs/heads/" + targetref

        self.__targetref = targetref
        self.__open_files: cabc.MutableMapping[
            tuple[int, pathlib.PurePosixPath], _ProcessWriter
        ] = weakref.WeakValueDictionary()

    def __enter__(self) -> cabc.Mapping[str, t.Any]:
        self.__updates = {}
        self.__old_sha = (
            self.__handler._git("rev-parse", self.__handler.revision)
            .decode("ascii")
            .strip()
        )
        if self.__handler._transaction is not None:
            raise RuntimeError("Another transaction is already open")
        self.__handler._transaction = self

        return self.__outer_context.__enter__()

    def __exit__(self, exc_type, exc_value, exc_trace):
        if exc_value is not None:
            return self.__outer_context.__exit__(
                exc_type, exc_value, exc_trace
            )
        try:
            LOGGER.debug("Creating updated tree structures")
            tree = self.__update_tree(
                self.__read_tree(self.__old_sha),
                self.__updates,
            )

            LOGGER.debug("Creating commit object with tree %s", tree)
            commit = self.__commit(tree)
            if self.__dry_run:
                LOGGER.debug("Not updating branch pointers (dry_run=True)")
                return None

            LOGGER.debug("Updating ref %r to %s", self.__targetref, commit)
            self.__update_target_ref(commit)

            if not self.__push:
                LOGGER.debug("Not pushing changes to remote (push=False)")
                return None

            LOGGER.debug("Pushing updated ref %r", self.__targetref)
            self.__push_updates("origin")
        finally:
            del self.__old_sha
            self.__handler._transaction = None
            self.__outer_context.__exit__(exc_type, exc_value, exc_trace)
        return None

    def record_update(self, path: pathlib.PurePosixPath, new_sha: str) -> None:
        """Record an updated blob in the current transaction.

        Parameters
        ----------
        path
            The path of the blob, relative to the root of the repository.
        new_sha
            The SHA sum (object name) of the new blob.
        """
        assert re.fullmatch("[0-9a-fA-F]+", new_sha)
        assert self.__handler._transaction is not None
        if path in self.__updates:
            LOGGER.warning(
                "Path changed twice in the same transaction: %s", path
            )
        self.__updates[path] = new_sha

    def record_pending_update(
        self, filename: pathlib.PurePosixPath, file: _ProcessWriter
    ) -> None:
        self.__open_files[id(file), filename] = file

    def __commit(self, tree: str) -> str:
        """Commit ``tree`` as child commit of ``__target_ref``."""
        commit_hash = (
            self.__handler._git(
                "commit-tree",
                tree,
                "-p",
                self.__old_sha,
                "-m",
                self.__commit_msg,
                env=self.__gitenv,
            )
            .strip()
            .decode("ascii")
        )
        LOGGER.debug("Created commit with hash %r", commit_hash)
        return commit_hash

    def __push_updates(self, remote: str) -> None:
        """Push the locally updated ``__target_ref`` to ``remote``."""
        self.__handler._git(
            "-c",
            f"remote.{remote}.mirror=false",
            "push",
            *self.__push_opts,
            "--",
            remote,
            self.__targetref,
        )

    def __read_tree(self, treeish: str) -> dict[str, _TreeEntry]:
        """Read a tree object from git into a dictionary."""
        treedata = self.__handler._git("ls-tree", treeish, "-z")
        tree: dict[str, _TreeEntry] = {}
        for line in treedata.decode("utf-8").split("\0"):
            if line == "":
                continue
            leaf = _TreeEntry.fromstring(line)
            tree[leaf.file] = leaf
        return tree

    def __resolve_head(self) -> str:
        """Resolve HEAD to a single symbolic name."""
        try:
            name = self.__handler._git(
                "rev-parse", "--symbolic-full-name", "HEAD", silent=True
            )
        except subprocess.CalledProcessError:
            name = b""
        return name.decode("utf-8").strip()

    def __update_target_ref(self, commit: str) -> None:
        """Update the local ``__target_ref``."""
        self.__handler._git(
            "update-ref",
            f"-m[{capellambse.__name__}] Commit by capellambse",
            self.__targetref,
            commit,
        )

    def __update_tree(
        self,
        old_tree: cabc.Mapping[str, _TreeEntry],
        updates: cabc.Mapping[pathlib.PurePosixPath, str],
    ) -> str:
        """Apply ``updates`` to ``old_tree`` and create a new tree object."""
        unclosed_files = 0
        for (_, filename), file in self.__open_files.items():
            if not file.closed:
                unclosed_files += 1
                LOGGER.warning("File is still open: %s", filename)
        if unclosed_files:
            LOGGER.critical(self.__unclosed_error, unclosed_files)
        tree = dict(old_tree)

        def groupkey(i: tuple[pathlib.PurePosixPath, str]) -> str:
            if len(i[0].parts) > 1:
                return i[0].parts[0]
            return ""

        for prefix, contents in itertools.groupby(
            sorted(updates.items(), key=groupkey), key=groupkey
        ):
            if not prefix:
                for f, h in contents:
                    tree[f.name] = _TreeEntry("100644", "blob", h, f.name)
            else:
                ent = tree.get(prefix)
                if ent is None or ent.type != "tree":
                    ent = _TreeEntry("040000", "tree", "", prefix)

                new_subtree = self.__update_tree(
                    self.__read_tree(ent.object) if ent.object else {},
                    {k.relative_to(prefix): v for k, v in contents},
                )
                tree[prefix] = ent._replace(object=new_subtree)

        tree_dump = "\0".join(i.tostring() for i in tree.values())
        tree_hash = (
            self.__handler._git(
                "mktree", "-z", input=tree_dump.encode("utf-8")
            )
            .decode("ascii")
            .strip()
        )
        LOGGER.debug("Created tree with hash %r", tree_hash)
        return tree_hash


class GitFileHandler(FileHandler):
    """File handler for ``git://`` and related protocols.

    Parameters
    ----------
    revision
        The Git revision to check out. Either a branch or tag name, a
        full ref name, or the object name (i.e. hash) of a commit-ish.
    username
        The user name for authentication with the Git remote.
    password
        The password for authentication with the Git remote.
    identity_file
        Authenticate against the remote with the private key in this
        file. Defaults to using SSH's algorithm for determining a
        suitable key. (SSH only, ignored otherwise)
    known_hosts_file
        An OpenSSH-style ``known_hosts`` file containing the public key
        of the remote server. (SSH only, ignored otherwise)
    disable_cache
        Wipe the local cache and create a fresh, new clone.
    update_cache
        Contact the remote and make sure that the local cache is up to
        date. Note that setting this to ``False`` does not necessarily
        inhibit all attempts to contact the remote; it just disables the
        initial "fetch" operation. Later operations may still require to
        access the server, for example to download Git-LFS files.
    shallow
        Make a shallow clone. This can drastically reduce network and
        disk usage when cloning large models, by only downloading the
        latest ``revision`` and not downloading the history that leads
        up to it.

        Note that, when this is set to ``True`` (the default), existing
        non-shallow caches will be made shallow. However, when it is set
        to ``False``, shallow caches will not be unshallowed.

    Attributes
    ----------
    cache_dir
        The path to the temporary work tree created by instances of this
        file handler.

    See Also
    --------
    capellambse.filehandler.FileHandler :
        Documentation of common parameters.
    """

    username: str
    password: str
    identity_file: str
    known_hosts_file: str
    cache_dir: pathlib.Path
    shallow: bool

    __fnz: object
    __has_lfs: bool
    __lfsfiles: dict[pathlib.PurePosixPath, bool]
    __repo: pathlib.Path

    def __init__(
        self,
        path: str | os.PathLike,
        revision: str = "HEAD",
        username: str = "",
        password: str = "",
        identity_file: str = "",
        known_hosts_file: str = "",
        disable_cache: bool = False,
        update_cache: bool = True,
        *,
        subdir: str | pathlib.PurePosixPath = "/",
        shallow: bool = True,
    ) -> None:
        super().__init__(path, subdir=subdir)
        self.disable_cache = disable_cache
        self.username = username
        self.password = password
        self.identity_file = identity_file
        self.known_hosts_file = known_hosts_file
        self.update_cache = update_cache
        self.shallow = shallow

        self.cache_dir = None  # type: ignore[assignment]
        self.__hash, self.revision = self.__resolve_remote_ref(revision)

        self.__init_cache_dir()
        self.__init_worktree()

        self._transaction: _GitTransaction | None = None
        try:
            self._git("lfs", "env", silent=True)
        except subprocess.CalledProcessError:
            LOGGER.debug("LFS not installed, disabling related functionality")
            self.__has_lfs = False
        else:
            LOGGER.debug("LFS support detected")
            self.__has_lfs = True
        self.__lfsfiles = {}

    def open(
        self,
        filename: str | pathlib.PurePosixPath,
        mode: t.Literal["r", "rb", "w", "wb"] = "rb",
    ) -> t.BinaryIO:
        path = capellambse.helpers.normalize_pure_path(
            filename, base=self.subdir
        )
        if "w" in mode:
            if self._transaction is None:
                raise TransactionClosedError(
                    "Writing to git requires a transaction"
                )
            return self.__open_writable(path)
        return self.__open_readable(path)

    def __open_readable(self, path: pathlib.PurePosixPath) -> t.BinaryIO:
        try:
            if self.__is_lfs(path):
                content = self.__open_from_lfs(path)
            else:
                content = self.__open_from_index(path)
        except subprocess.CalledProcessError as err:
            stderr = err.stderr.decode("utf-8")
            if str(path) in stderr.splitlines()[0]:
                raise FileNotFoundError(stderr) from err
            raise
        return io.BytesIO(content)

    def __open_writable(self, path: pathlib.PurePosixPath) -> t.BinaryIO:
        assert self._transaction is not None

        cls: type[_WritableLFSFile] | type[_WritableIndexFile]
        if self.__is_lfs(path):
            cls = _WritableLFSFile
        else:
            cls = _WritableIndexFile
        file = cls(
            cb=functools.partial(self._transaction.record_update, path),
            cwd=self.cache_dir,
            env=self.__get_git_env(),
            filename=path,
        )
        self._transaction.record_pending_update(path, file)
        return file

    def get_model_info(self) -> modelinfo.ModelInfo:
        def revparse(*args: str) -> str:
            return (
                self._git("rev-parse", *args, silent=True)
                .decode("utf-8", errors="surrogateescape")
                .strip()
            )

        title = str(self.path).rsplit("/", maxsplit=1)[-1]
        if title.endswith(".git"):
            title = title[: -len(".git")]

        return modelinfo.ModelInfo(
            branch=revparse("--abbrev-ref", self.revision),
            title=title,
            url=str(self.path),
            revision=self.revision,
            rev_hash=revparse(self.revision),
        )

    def write_transaction(
        self, **kw: t.Any
    ) -> t.ContextManager[cabc.Mapping[str, t.Any]]:
        """See _GitTransaction for the actual docstring."""
        return _GitTransaction(super().write_transaction, self, **kw)

    write_transaction.__doc__ = _GitTransaction.__init__.__doc__

    @staticmethod
    def __cleanup_worktree(
        repo_root: pathlib.Path, worktree: pathlib.Path, /
    ) -> None:
        LOGGER.debug("Removing worktree at %s", worktree)
        subprocess.run(
            ["git", "worktree", "remove", "-f", str(worktree)],
            check=True,
            cwd=repo_root,
        )

    def __get_git_env(self) -> dict[str, str]:
        git_env = os.environ.copy()
        if not os.environ.get("GIT_ASKPASS"):
            path_to_askpass = (
                pathlib.Path(__file__).parent / "git_askpass.py"
            ).absolute()

            git_env["GIT_ASKPASS"] = str(path_to_askpass)

            try:
                os.chmod(path_to_askpass, 0o755)
            except OSError:
                LOGGER.info(
                    "Setting permission 755 for GIT_ASKPASS file failed"
                )

        if self.username and self.password:
            git_env["GIT_USERNAME"] = self.username
            git_env["GIT_PASSWORD"] = self.password

        if self.identity_file and self.known_hosts_file:
            ssh_command = [
                "ssh",
                "-i",
                self.identity_file,
                f"-oUserKnownHostsFile={self.known_hosts_file}",
            ]
            git_env["GIT_SSH_COMMAND"] = shlex.join(ssh_command)

        return git_env

    def __init_cache_dir(self) -> None:
        if str(self.path).startswith("file://"):
            self.__init_cache_dir_local()
        else:
            self.__init_cache_dir_remote()

    def __init_cache_dir_local(self) -> None:
        parts = urllib.parse.urlparse(str(self.path))
        if parts.netloc and parts.netloc != "localhost":
            raise ValueError(f"Unsupported file:// URL netloc: {parts.netloc}")

        path = pathlib.Path(urllib.parse.unquote(parts.path))
        if isinstance(path, pathlib.WindowsPath):
            path = pathlib.Path(str(path)[1:])
        assert path.is_absolute()
        self.__repo = self.cache_dir = path.resolve()

    def __init_cache_dir_remote(self) -> None:
        slug_pattern = '[\x00-\x1f\x7f"*/:<>?\\|]+'
        path_hash = hashlib.sha256(
            str(self.path).encode("utf-8", errors="surrogatepass")
        ).hexdigest()
        path_slug = re.sub(slug_pattern, "-", str(self.path))
        self.__repo = self.cache_dir = pathlib.Path(
            capellambse.dirs.user_cache_dir,
            "models",
            path_hash,
            path_slug,
        )

        if self.cache_dir.exists() and self.disable_cache:
            shutil.rmtree(str(self.cache_dir))

        update_cache = self.update_cache
        if not (self.cache_dir / "config").exists():
            self.cache_dir.mkdir(parents=True, exist_ok=True)
            LOGGER.debug("Creating a new git repo in %s", self.cache_dir)
            self._git("-c", "init.defaultBranch=master", "init", "--bare")
            self._git("remote", "add", "--mirror=fetch", "origin", self.path)
            update_cache = True

        if update_cache:
            LOGGER.debug("Updating cache at %s", self.cache_dir)
            shallow_opts: tuple[str, ...] = ()
            if self.shallow:
                fetchspec = f"+{self.revision}"
                if not _git_object_name.search(self.revision):
                    fetchspec += f":{self.revision}"
                shallow_opts = ("--depth=1", fetchspec)
            self._git("fetch", self.path, *shallow_opts)

    def __resolve_remote_ref(self, ref: str) -> tuple[str, str]:
        """Resolve the given ``ref`` on the remote."""
        LOGGER.debug("Resolving ref %r on remote %s", ref, self.path)
        listing = self.__git_nolock(
            "ls-remote", self.path, ref, encoding="utf-8"
        )
        assert isinstance(listing, str)
        if not listing:
            if not _git_object_name.search(ref):
                raise ValueError(f"Ref does not exist on remote: {ref}")
            LOGGER.debug("Ref %r not found, assuming object name", ref)
            return ref, ref
        refs = [i.split("\t") for i in listing.strip().split("\n")]
        if len(refs) > 1:
            raise ValueError(
                f"Ambiguous ref name {ref}, found {len(refs)}:"
                f" {', '.join(i[1] for i in refs)}"
            )

        hash, refname = refs[0]
        LOGGER.debug(
            "Resolved ref %r as remote ref %r (%s)", ref, refname, hash
        )
        return hash, refname

    def __init_worktree(self) -> None:
        worktree = pathlib.Path(
            tempfile.mkdtemp(None, f"capellambse-{os.getpid()}-")
        )
        LOGGER.debug("Setting up a worktree at %s", worktree)
        try:
            self._git(
                "worktree",
                "add",
                "--detach",
                "--no-checkout",
                worktree,
                self.__hash,
            )
        except:
            os.rmdir(worktree)
            raise
        self.__repo = self.cache_dir
        self.cache_dir = worktree

        self.__fnz = weakref.finalize(  # pylint: disable=unused-private-member
            self, self.__cleanup_worktree, self.__repo, worktree
        )

        self._git("reset", "--mixed", self.revision)

    def __is_lfs(self, path: pathlib.PurePosixPath) -> bool:
        """Check whether a file uses LFS or not.

        As a side effect, this method will raise an exception if
        ``__init__`` has detected that ``git-lfs`` is not installed and
        the ``path`` is determined to be an LFS file.
        """
        try:
            return self.__lfsfiles[path]
        except KeyError:
            pass

        attrs = self._git("check-attr", "--all", "--cached", "-z", "--", path)
        for _, attr, value in capellambse.helpers.ntuples(
            3, attrs.split(b"\0")
        ):
            if attr == b"filter" and value == b"lfs":
                break
        else:
            self.__lfsfiles[path] = False
            return False

        self.__lfsfiles[path] = True
        if not self.__has_lfs:
            raise RuntimeError(
                f"Cannot open LFS file, git-lfs is not installed: {path}"
            )
        return True

    def __open_from_index(self, filename: pathlib.PurePosixPath) -> bytes:
        return self._git("cat-file", "blob", f"{self.revision}:{filename}")

    def __open_from_lfs(self, filename: pathlib.PurePosixPath) -> bytes:
        lfsinfo = self.__open_from_index(filename)
        return self._git("lfs", "smudge", "--", filename, input=lfsinfo)

    @t.overload
    def _git(
        self,
        *cmd: t.Any,
        encoding: str,
        env: cabc.Mapping[str, str] | None = ...,
        silent: bool = ...,
        **kw: t.Any,
    ) -> str:
        ...

    @t.overload
    def _git(
        self,
        *cmd: t.Any,
        encoding: None = ...,
        env: cabc.Mapping[str, str] | None = ...,
        silent: bool = ...,
        **kw: t.Any,
    ) -> bytes:
        ...

    def _git(
        self,
        *cmd: t.Any,
        env: cabc.Mapping[str, str] | None = None,
        silent: bool = False,
        **kw: t.Any,
    ) -> bytes | str:
        with capellambse.helpers.flock(self.__repo / "capellambse.lock"):
            return self.__git_nolock(*cmd, env=env, silent=silent, **kw)

    def __git_nolock(
        self,
        *cmd: t.Any,
        env: cabc.Mapping[str, str] | None = None,
        silent: bool = False,
        **kw: t.Any,
    ) -> bytes | str:
        LOGGER.debug("Running command %s", cmd)
        returncode = 0
        stderr = None
        try:
            proc = subprocess.run(
                ["git"] + [str(i) for i in cmd],
                capture_output=True,
                check=True,
                cwd=self.cache_dir,
                env={**self.__get_git_env(), **(env or {})},
                **kw,
            )
            returncode = proc.returncode
            stderr = proc.stderr
            return proc.stdout
        except subprocess.CalledProcessError as err:
            returncode = err.returncode
            stderr = err.stderr
            raise
        finally:
            if returncode != 0 and not silent:
                level = logging.ERROR
            else:
                level = logging.DEBUG

            if stderr:
                if isinstance(stderr, bytes):
                    stderr = stderr.decode("utf-8")

                for line in stderr.splitlines():
                    LOGGER.getChild("git").log(level, "%s", line)
            LOGGER.log(level, "Exit status: %d", returncode)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Support for YAML-based declarative modelling.

A YAML-based approach to describing how to create and modify
``capellambse`` compatible models.

For an in-depth explanation, please refer to the :ref:`full
documentation about declarative modelling <declarative-modelling>`.
"""
from __future__ import annotations

__all__ = [
    "Promise",
    "UUIDReference",
    "UnfulfilledPromisesError",
    "YDMDumper",
    "YDMLoader",
    "apply",
    "dump",
    "load",
]

import collections
import collections.abc as cabc
import contextlib
import dataclasses
import os
import sys
import typing as t
import warnings

import yaml

import capellambse
from capellambse import helpers
from capellambse.model import common

FileOrPath = t.Union[t.IO[str], str, os.PathLike[t.Any]]
_FutureAction = dict[str, t.Any]
_OperatorResult = tuple[
    "Promise",
    t.Union[capellambse.ModelObject, _FutureAction],
]


def dump(instructions: cabc.Sequence[cabc.Mapping[str, t.Any]]) -> str:
    """Dump an instruction stream to YAML."""
    return yaml.dump(instructions, Dumper=YDMDumper)


def load(file: FileOrPath) -> list[dict[str, t.Any]]:
    """Load an instruction stream from a YAML file.

    Parameters
    ----------
    file
        An open file-like object, or a path or PathLike pointing to such
        a file. Files are expected to use UTF-8 encoding.
    """
    if hasattr(file, "read"):
        file = t.cast(t.IO[str], file)
        ctx: t.ContextManager[t.IO[str]] = contextlib.nullcontext(file)
    else:
        assert not isinstance(file, t.IO)
        ctx = open(file, encoding="utf-8")

    with ctx as opened_file:
        return yaml.load(opened_file, Loader=YDMLoader)


def apply(model: capellambse.MelodyModel, file: FileOrPath) -> None:
    """Apply a declarative modelling file to the given model.

    Parameters
    ----------
    model
        The model to apply the instructions to.
    file
        An open file-like object to read YAML instructions from, or a
        path to such a file. Files will be read with UTF-8 encoding.

        The full format of these files is documented in the
        :ref:`section about declarative modelling
        <declarative-modelling>`.

    Notes
    -----
    This function is not transactional: If an exception occurs during
    this function, the model will be left partially modified, with no
    reliable way to know how much of the YAML input has been consumed.
    It is therefore advised to reload or discard the model immediately
    in these cases, to avoid working with an inconsistent state.

    Even though the YAML layout suggests linear execution from top to
    bottom, the actual order in which modifications are executed is
    implementation defined. This is necessary to support promises with
    ``!promise``, but reorderings are still possible even if no promises
    are used in an input document.
    """
    instructions = collections.deque(load(file))
    promises = dict[Promise, capellambse.ModelObject]()
    deferred = collections.defaultdict[Promise, list[_FutureAction]](list)

    while instructions:
        instruction = instructions.popleft()
        parent = instruction.pop("parent")
        if isinstance(parent, UUIDReference):
            parent = model.by_uuid(parent.uuid)
        for op_type, apply_op in _OPERATIONS.items():
            try:
                op = instruction.pop(op_type)
            except KeyError:
                continue
            for promise, outcome in apply_op(promises, parent, op):
                if isinstance(outcome, dict):
                    deferred[promise].append(outcome)
                else:
                    if promise in promises:
                        raise ValueError(
                            f"promise_id defined twice: {promise.identifier}"
                        )
                    promises[promise] = outcome
                    instructions.extend(deferred.pop(promise, ()))
        if instruction:
            keys = ", ".join(instruction)
            raise ValueError(f"Unrecognized keys in instruction: {keys}")
    if deferred:
        raise UnfulfilledPromisesError(frozenset(deferred))


def _operate_create(
    promises: dict[Promise, capellambse.ModelObject],
    parent: capellambse.ModelObject,
    creations: dict[str, t.Any],
) -> cabc.Generator[_OperatorResult, t.Any, None]:
    warnings.warn("Use 'extend' instead of 'create' in declarative YAML")
    yield from _operate_extend(promises, parent, creations)


def _operate_extend(
    promises: dict[Promise, capellambse.ModelObject],
    parent: capellambse.ModelObject,
    extensions: dict[str, t.Any],
) -> cabc.Generator[_OperatorResult, t.Any, None]:
    for attr, value in extensions.items():
        if not isinstance(value, cabc.Iterable):
            raise TypeError("values below `extend:*:` must be lists")

        yield from _create_complex_objects(promises, parent, attr, value)


def _operate_delete(
    promises: dict[Promise, capellambse.ModelObject],
    parent: capellambse.ModelObject,
    deletions: dict[str, t.Any],
) -> cabc.Iterable[_OperatorResult]:
    del promises

    for attr, objs in deletions.items():
        try:
            target = getattr(parent, attr)
        except AttributeError:
            raise TypeError(
                "Cannot delete object:"
                f" {type(parent).__name__} has no attribute {attr!r}"
            ) from None
        if not isinstance(target, common.ElementList) or not isinstance(
            objs, list
        ):
            delattr(parent, attr)
            continue
        if not isinstance(target, common.ElementListCouplingMixin):
            raise TypeError(
                "Cannot delete object:"
                f" {type(parent).__name__}.{attr} is not model-coupled"
            )
        uuids = list(t.cast(cabc.Iterable[str], target.by_uuid))
        for obj in objs:
            if isinstance(obj, UUIDReference):
                obj = obj.uuid
            if not isinstance(obj, str):
                raise TypeError("Values in `delete:*:` must be UUIDs")
            try:
                idx = uuids.index(obj)
            except IndexError:
                puuid = getattr(parent, "uuid", None)
                raise ValueError(
                    f"No object with UUID {obj!r} in {attr!r} of {puuid!r}"
                ) from None
            del uuids[idx]
            del target[idx]

    return ()


def _operate_modify(
    promises: dict[Promise, capellambse.ModelObject],
    parent: capellambse.ModelObject,
    modifications: dict[str, t.Any],
) -> cabc.Generator[_OperatorResult, t.Any, None]:
    for attr, value in modifications.items():
        if isinstance(value, (list, Promise, UUIDReference)):
            try:
                value = _resolve(promises, parent, value)
            except _UnresolvablePromise as p:
                yield p.args[0], {"parent": parent, "modify": {attr: value}}
                continue

        if isinstance(value, list):
            delattr(parent, attr)
            yield from _create_complex_objects(promises, parent, attr, value)
        elif isinstance(value, dict):
            obj = getattr(parent, attr)
            for k, v in value.items():
                setattr(obj, k, v)
        else:
            setattr(parent, attr, value)


def _resolve(
    promises: dict[Promise, capellambse.ModelObject],
    parent: capellambse.ModelObject,
    value: t.Any,
) -> t.Any:
    if isinstance(value, Promise):
        try:
            return promises[value]
        except KeyError:
            raise _UnresolvablePromise(value) from None
    elif isinstance(value, UUIDReference):
        return parent._model.by_uuid(value.uuid)
    elif isinstance(value, list):
        for i, v in enumerate(value):
            newv = _resolve(promises, parent, v)
            if newv is not v:
                value[i] = newv
    return value


class _UnresolvablePromise(BaseException):
    pass


_OPERATIONS = collections.OrderedDict(
    (
        ("create", _operate_create),
        ("extend", _operate_extend),
        ("modify", _operate_modify),
        ("delete", _operate_delete),
    )
)


def _create_complex_objects(
    promises: dict[Promise, capellambse.ModelObject],
    parent: capellambse.ModelObject,
    attr: str,
    objs: cabc.Iterable[dict[str, t.Any] | Promise | str],
) -> cabc.Generator[_OperatorResult, t.Any, None]:
    try:
        target = getattr(parent, attr)
    except AttributeError:
        raise TypeError(
            "Cannot create object:"
            f" {type(parent).__name__} has no attribute {attr!r}"
        ) from None
    if not isinstance(target, common.ElementList):
        raise TypeError(
            "Cannot create object:"
            f" {type(parent).__name__}.{attr} is not a list"
        )
    if not isinstance(target, common.ElementListCouplingMixin):
        raise TypeError(
            "Cannot create object:"
            f" {type(parent).__name__}.{attr} is not model-coupled"
        )
    for child in objs:
        if isinstance(
            child, (common.GenericElement, list, Promise, UUIDReference)
        ):
            try:
                obj = _resolve(promises, parent, child)
            except _UnresolvablePromise as p:
                yield p.args[0], {"parent": parent, "extend": {attr: [child]}}
            else:
                target.append(obj)
        elif isinstance(child, str):
            target.create_singleattr(child)
        else:
            yield from _create_complex_object(
                promises, parent, attr, target, child
            )


def _create_complex_object(
    promises: dict[Promise, capellambse.ModelObject],
    parent: capellambse.ModelObject,
    attr: str,
    target: common.ElementListCouplingMixin,
    obj_desc: dict[str, t.Any],
) -> cabc.Generator[_OperatorResult, t.Any, None]:
    try:
        promise: str | Promise | None = obj_desc.pop("promise_id")
    except KeyError:
        promise = None

    complex_attrs = dict[str, t.Any]()
    simple_attrs = dict[str, t.Any]()
    type_hint = tuple[str, ...]()
    try:
        for k, v in obj_desc.items():
            if k == "_type":
                type_hint = (v,)
            elif isinstance(v, cabc.Iterable) and not isinstance(v, str):
                complex_attrs[k] = v
            else:
                simple_attrs[k] = _resolve(promises, parent, v)
    except _UnresolvablePromise as p:
        obj_desc["promise_id"] = promise
        yield (p.args[0], {"parent": parent, "extend": {attr: [obj_desc]}})
        return
    assert isinstance(target, common.ElementListCouplingMixin)
    obj = target.create(*type_hint, **simple_attrs)
    if promise is not None:
        if isinstance(promise, str):
            promise = Promise(promise)
        yield (promise, obj)
    for subkey, subval in complex_attrs.items():
        if isinstance(subval, cabc.Iterable):
            yield from _create_complex_objects(promises, obj, subkey, subval)
        else:
            raise TypeError(f"Expected list, got {type(subval).__name__}")


@dataclasses.dataclass(frozen=True)
class Promise:
    """References a model object that will be created later."""

    identifier: str


class UnfulfilledPromisesError(RuntimeError):
    """A promise could not be fulfilled.

    This exception is raised when a promise is referenced via
    ``!promise``, but it is never fulfilled by declaring an object with
    the same ``promise_id``.
    """

    def __str__(self) -> str:
        if (
            len(self.args) == 1
            and isinstance(self.args[0], cabc.Iterable)
            and not isinstance(self.args[0], str)
        ):
            return ", ".join(i.identifier for i in self.args[0])
        return super().__str__()


@dataclasses.dataclass(frozen=True)
class UUIDReference:
    """References a model object by its UUID."""

    uuid: helpers.UUIDString

    def __post_init__(self) -> None:
        if not helpers.is_uuid_string(self.uuid):
            raise ValueError(f"Malformed `!uuid`: {self.uuid!r}")


class YDMDumper(yaml.SafeDumper):
    """A YAML dumper with extensions for declarative modelling."""

    def represent_promise(self, data: t.Any) -> yaml.Node:
        assert isinstance(data, Promise)
        return self.represent_scalar("!promise", data.identifier)

    def represent_uuidref(self, data: t.Any) -> yaml.Node:
        assert isinstance(data, UUIDReference)
        return self.represent_scalar("!uuid", data.uuid)


YDMDumper.add_representer(Promise, YDMDumper.represent_promise)
YDMDumper.add_representer(UUIDReference, YDMDumper.represent_uuidref)


class YDMLoader(yaml.SafeLoader):
    """A YAML loader with extensions for declarative modelling."""

    def construct_promise(self, node: yaml.Node) -> Promise:
        if not isinstance(node, yaml.ScalarNode):
            raise TypeError("!promise only accepts scalar nodes")
        data = self.construct_scalar(node)
        if not isinstance(data, str):
            raise TypeError("!promise only accepts string scalars")
        return Promise(data)

    def construct_uuidref(self, node: yaml.Node) -> UUIDReference:
        if not isinstance(node, yaml.ScalarNode):
            raise TypeError("!uuid only accepts scalar nodes")
        data = self.construct_scalar(node)
        if not helpers.is_uuid_string(data):
            raise ValueError(f"Not a well-formed UUID string: {data}")
        return UUIDReference(data)


YDMLoader.add_constructor("!promise", YDMLoader.construct_promise)
YDMLoader.add_constructor("!uuid", YDMLoader.construct_uuidref)


try:
    import click
except ImportError:

    def _main() -> None:
        """Display a dependency error."""
        print("Error: Please install 'click' and retry", file=sys.stderr)
        raise SystemExit(1)

else:

    @click.command()
    @click.option("-m", "--model", type=capellambse.ModelCLI(), required=True)
    @click.argument("file", type=click.File("r"))
    def _main(model: capellambse.MelodyModel, file: t.IO[str]) -> None:
        """Apply a declarative modelling YAML file to a model."""
        apply(model, file)
        model.save()


if __name__ == "__main__":
    _main()



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

import capellambse

HIDDEN_UUID = "957c5799-1d4a-4ac0-b5de-33a65bf1519c"


def test_diagram_nodes_only_include_visible_elements(
    session_shared_model: capellambse.MelodyModel,
):
    diagram = session_shared_model.diagrams.by_name(
        "[LAB] Hidden Wizzard Education"
    )

    assert HIDDEN_UUID not in diagram.nodes.by_uuid
